# -*- coding: utf-8 -*-
import pytest

from schematics.models import Model
from schematics.types import StringType, LongType, IntType, MD5Type
from schematics.types.compound import ModelType, DictType, ListType
from schematics.types.serializable import serializable
from schematics.transforms import blacklist, whitelist, wholelist, export_loop

import six
from six import iteritems
try:
    unicode #PY2
except:
    import codecs
    unicode = str #PY3

def test_serializable():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    location_US = Location({"country_code": "US"})

    assert location_US.country_name == "United States"

    d = location_US.serialize()
    assert d == {"country_code": "US", "country_name": "United States"}

    d = location_US.to_native()
    assert d == {"country_code": u"US", "country_name": "United States"}

    location_IS = Location({"country_code": "IS"})

    assert location_IS.country_name == "Unknown"

    d = location_IS.serialize()
    assert d == {"country_code": "IS", "country_name": "Unknown"}

    d = location_IS.to_native()
    assert d == {"country_code": "IS", "country_name": "Unknown"}


def test_serializable_to_native():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    loc = Location({'country_code': 'US'})

    d = loc.to_native()
    assert d == {'country_code': 'US', 'country_name': 'United States'}


def test_serializable_with_serializable_name():
    class Location(Model):
        country_code = StringType(serialized_name="cc")

        @serializable(serialized_name="cn")
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    location_US = Location({"cc": "US"})

    assert location_US.country_name == "United States"

    d = location_US.serialize()
    assert d == {"cc": "US", "cn": "United States"}


def test_serializable_with_custom_serializable_class():
    class PlayerIdType(LongType):

        def to_primitive(self, value, context=None):
            return unicode(value)

    class Player(Model):
        id = LongType()

        @serializable(type=PlayerIdType())
        def player_id(self):
            return self.id

    player = Player({"id": 1})

    assert player.id == 1
    assert player.player_id == 1

    d = player.serialize()
    assert d == {"id": 1, "player_id": "1"}


def test_serializable_with_model():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel(dict(level=self.total_points * 2, title="Best"))

    player = Player({"total_points": 2})

    assert player.xp_level.level == 4

    d = player.serialize()
    assert d == {"total_points": 2, "xp_level": {"level": 4, "title": "Best"}}


def test_serializable_with_model_to_native():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel(dict(level=self.total_points * 2, title="Best"))

    player = Player({"total_points": 2})

    assert player.xp_level.level == 4

    d = player.to_native()
    assert d == {"total_points": 2, "xp_level": {"level": 4, "title": "Best"}}


def test_serializable_with_model_when_None():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return None if not self.total_points else ExperienceLevel()

    player = Player({"total_points": 0})

    assert player.xp_level is None

    d = player.serialize()
    assert d == {"total_points": 0, "xp_level": None}


def test_serializable_with_model_hide_None():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel), serialize_when_none=False)
        def xp_level(self):
            return None if not self.total_points else ExperienceLevel()

    player = Player({"total_points": 0})

    assert player.xp_level is None

    d = player.serialize()
    assert d == {"total_points": 0}


def test_serializable_with_embedded_models_and_list():
    class Question(Model):
        id = LongType()

    class QuestionPack(Model):
        id = LongType()
        questions = ListType(ModelType(Question))

    class Game(Model):
        id = StringType()
        question_pack = ModelType(QuestionPack)

    q1 = Question({"id": 1})
    q2 = Question({"id": 2})

    game = Game({
        "id": "1",
        "question_pack": {
            "id": 2,
            "questions": [q1, q2]
        }
    })

    assert game.question_pack.questions[0] == q1
    assert game.question_pack.questions[1] == q2

    d = game.serialize()

    assert d == {
        "id": "1",
        "question_pack": {
            "id": 2,
            "questions": [
                {
                    "id": 1,
                },
                {
                    "id": 2,
                },
            ]
        }
    }


def test_serializable_with_embedded_models():
    class ExperienceLevel(Model):
        level = IntType()
        stars = IntType()

        @classmethod
        def from_total_points(cls, total_points, category_slug):
            return cls(dict(level=total_points * 2, stars=total_points))

    class CategoryStatsInfo(Model):
        category_slug = StringType()
        total_points = IntType(default=0)

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel.from_total_points(self.total_points, self.category_slug)

    class PlayerInfo(Model):
        id = LongType()
        display_name = StringType()

    class PlayerCategoryInfo(PlayerInfo):
        categories = DictType(ModelType(CategoryStatsInfo))

    info = PlayerCategoryInfo(dict(
        id="1",
        display_name="John Doe",
        categories={
            "math": {
                "category_slug": "math",
                "total_points": 1
            }
        }
    ))

    assert info.categories["math"].xp_level.level == 2
    assert info.categories["math"].xp_level.stars == 1

    d = info.serialize()
    assert d == {
        "id": 1,
        "display_name": "John Doe",
        "categories": {
            "math": {
                "category_slug": "math",
                "total_points": 1,
                "xp_level": {
                    "level": 2,
                    "stars": 1,
                }
            }
        }
    }


def test_serializable_works_with_inheritance():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    class LocationWithCity(Location):
        city_code = StringType()

        @serializable
        def city_name(self):
            return "Oklahoma" if self.city_code == "OK" else "Unknown"

    location = LocationWithCity(dict(country_code="US", city_code="OK"))

    assert location.country_code == "US"
    assert location.country_name == "United States"
    assert location.city_code == "OK"
    assert location.city_name == "Oklahoma"

    d = location.serialize()
    assert d == {
        "country_code": "US",
        "country_name": "United States",
        "city_code": "OK",
        "city_name": "Oklahoma",
    }


def test_serialize_with_complex_types():
    class QuestionResource(Model):
        url = StringType()

    class Question(Model):
        question_id = StringType(required=True)
        resources = DictType(ListType(ModelType(QuestionResource)))

    q = Question(dict(
        question_id="1",
        resources={
            "pictures": [{
                "url": "http://www.mbl.is",
            }]
        }
    ))

    d = q.serialize()
    assert d == dict(
        question_id="1",
        resources={
            "pictures": [{
                "url": "http://www.mbl.is",
            }]
        }
    )

    q_with_no_resources = Question(dict(
        question_id="1"
    ))

    d = q_with_no_resources.serialize()
    assert d == dict(
        question_id="1",
        resources=None,
    )


def test_field_with_serialize_when_none():
    class Question(Model):
        id = StringType()
        question = StringType()
        resources = DictType(StringType, serialize_when_none=False)

    q = Question(dict(id=1, question="Who's the man?"))

    d = q.serialize()
    assert d == {
        "id": "1",
        "question": "Who's the man?",
    }

    q = Question(dict(id=1, question="Who's the man?", resources={"A": "B"}))

    d = q.serialize()
    assert d == {
        "id": "1",
        "question": "Who's the man?",
        "resources": {"A": "B"}
    }


def test_field_with_serialize_when_none_on_outer_only():
    class M(Model):
        listfield = ListType(StringType(serialize_when_none=True), serialize_when_none=False)
        dictfield = DictType(StringType(serialize_when_none=True), serialize_when_none=False)
    obj = M()
    obj.listfield = [None]
    obj.dictfield = {'foo': None}
    assert obj.serialize() == {'listfield': [None], 'dictfield': {'foo': None}}


def test_field_with_serialize_when_none_on_inner_only():
    class M(Model):
        listfield = ListType(StringType(serialize_when_none=False), serialize_when_none=True)
        dictfield = DictType(StringType(serialize_when_none=False), serialize_when_none=True)
    obj = M()
    obj.listfield = [None]
    obj.dictfield = {'foo': None}
    assert obj.serialize() == {'listfield': [], 'dictfield': {}}


def test_set_serialize_when_none_on_whole_model():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType)

        class Options:
            serialize_when_none = False

    q = Question(dict(id=1))

    d = q.serialize()
    assert d == {"id": "1"}


def test_possible_to_override_model_wide_serialize_when_none():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType)

        class Options:
            serialize_when_none = False

    class StrictQuestion(Question):
        strictness = IntType()

        class Options:
            serialize_when_none = True

    q = StrictQuestion(dict(id=1))

    d = q.serialize()
    assert d == {"id": "1", "question": None, "resources": None, "strictness": None}


def test_possible_to_override_model_wide_settings_per_field():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType, serialize_when_none=True)

        class Options:
            serialize_when_none = False

    q = Question(dict(id=1))

    d = q.serialize()
    assert d == {"id": "1", "resources": None}


def test_complex_types_hiding_after_apply_role_leaves_it_empty():
    class QuestionResource(Model):
        name = StringType()
        url = StringType()

        class Options:
            serialize_when_none = False
            roles = {'public': whitelist('name')}

    class Question(Model):
        question_id = StringType(required=True)
        resources = DictType(ListType(ModelType(QuestionResource)))

        class Options:
            serialize_when_none = False
            roles = {'public': whitelist('question_id', 'resources')}

    q = Question(dict(
        question_id="1",
        resources={
            "pictures": [{
                "url": "http://www.mbl.is",
            }]
        }
    ))

    d = q.serialize('public')
    assert d == {'question_id': '1'}


def test_serialize_none_fields_if_field_says_so():
    class TestModel(Model):
        inst_id = StringType(required=True, serialize_when_none=True)

    q = TestModel({'inst_id': 1})

    d = export_loop(TestModel, q, lambda field, value: None)
    assert d == {'inst_id': None}


def test_serialize_none_fields_if_export_loop_says_so():
    class TestModel(Model):
        inst_id = StringType(required=True, serialize_when_none=False)

    q = TestModel({'inst_id': 1})

    d = export_loop(TestModel, q, lambda field, value: None, print_none=True)
    assert d == {'inst_id': None}


def test_serialize_print_none_always_gets_you_something():
    class TestModel(Model):
        pass

    q = TestModel()

    d = export_loop(TestModel, q, lambda field, value: None, print_none=True)
    assert d == {}


def test_roles_work_with_subclassing():
    class Address(Model):
        private_key = StringType()
        city = StringType()

        class Options:
            roles = {'public': blacklist('private_key')}

    class AddressWithPostalCode(Address):
        postal_code = IntType()

    a = AddressWithPostalCode(dict(
        postal_code=101,
        city=u"Reykjavík",
        private_key="secret"
    ))

    d = a.serialize(role="public")
    assert d == {
        "city": u"Reykjavík",
        "postal_code": 101,
    }


def test_role_propagate():
    class Address(Model):
        city = StringType()

        class Options:
            roles = {'public': whitelist('city')}

    class User(Model):
        name = StringType(required=True)
        password = StringType()
        addresses = ListType(ModelType(Address))

        class Options:
            roles = {'public': whitelist('name')}

    model = User({'name': 'a', 'addresses': [{'city': 'gotham'}]})
    assert model.addresses[0].city == 'gotham'

    d = model.serialize(role="public")
    assert d == {
        "name": "a",
    }


def test_fails_if_role_is_not_found():
    class Player(Model):
        id = StringType()

    p = Player(dict(id="1"))

    with pytest.raises(ValueError):
        p.serialize(role="public")


def test_doesnt_fail_if_role_isnt_found_on_embedded_models():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

        class Options:
            roles = {
                "public": wholelist(),
            }

    class Player(Model):
        id = StringType()
        secret = StringType()

        xp_level = ModelType(ExperienceLevel)

        class Options:
            roles = {
                "public": blacklist("secret")
            }

    p = Player(dict(
        id="1",
        secret="super_secret",
        xp_level={
            "level": 1,
            "title": "Starter"
        }
    ))

    d = p.serialize(role="public")
    assert d == {
        "id": "1",
        "xp_level": {
            "level": 1,
            "title": "Starter",
        }
    }


def test_doesnt_fail_serialize_when_none_on_whole_model_with_roles():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType)

        class Options:
            serialize_when_none = False
            roles = {
                "public": whitelist("id"),
            }

    q = Question({"id": "1"})

    d = q.serialize(role="public")
    assert d == {"id": "1"}


def test_uses_roles_on_embedded_models_if_found():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

        class Options:
            roles = {
                "public": blacklist("title")
            }

    class Player(Model):
        id = StringType()
        secret = StringType()

        xp_level = ModelType(ExperienceLevel)

        class Options:
            roles = {
                "public": blacklist("secret")
            }

    p = Player(dict(
        id="1",
        secret="super_secret",
        xp_level={
            "level": 1,
            "title": "Starter"
        }
    ))

    d = p.serialize(role="public")
    assert d == {
        "id": "1",
        "xp_level": {
            "level": 1,
        }
    }


def test_serializable_with_dict_and_roles():
    class Player(Model):
        id = LongType()
        display_name = StringType()

        class Options:
            roles = {
                "public": blacklist("id")
            }

    class Game(Model):
        id = StringType()
        result = IntType()
        players = DictType(ModelType(Player), coerce_key=lambda k: int(k))

        class Options:
            roles = {
                "public": blacklist("result")
            }

    p1 = Player({"id": 1, "display_name": "A"})
    p2 = Player({"id": 2, "display_name": "B"})

    game = Game({
        "id": "1",
        "players": {
            1: p1,
            2: p2
        }
    })

    assert game.players[1] == p1
    assert game.players[2] == p2

    d = game.serialize(role="public")

    assert d == {
        "id": "1",
        "players": {
            1: {
                "display_name": "A"
            },
            2: {
                "display_name": "B"
            },
        }
    }


def test_serializable_with_list_and_roles():
    class Player(Model):
        id = LongType()
        display_name = StringType()

        class Options:
            roles = {
                "public": blacklist("id")
            }

    class Game(Model):
        id = StringType()
        result = IntType()
        players = ListType(ModelType(Player))

        class Options:
            roles = {
                "public": blacklist("result")
            }

    p1 = Player({"id": 1, "display_name": "A"})
    p2 = Player({"id": 2, "display_name": "B"})

    game = Game({
        "id": "1",
        "players": [p1, p2]
    })

    assert game.players[0] == p1
    assert game.players[1] == p2

    d = game.serialize(role="public")

    assert d == {
        "id": "1",
        "players": [
            {
                "display_name": "A",
            },
            {
                "display_name": "B",
            },
        ]
    }


def test_role_set_operations():

    protected = whitelist('email', 'password')
    all_fields = whitelist('id', 'name') + protected

    def count(n):
        while True:
            yield n
            n += 1

    class User(Model):
        id = IntType(default=six.next(count(42)))
        name = StringType()
        email = StringType()
        password = StringType()

        class Options:
            roles = {
                'create': all_fields - ['id'],
                'public': all_fields - ['password'],
                'nospam': blacklist('password') + blacklist('email'),
                'empty': whitelist(),
                'everything': blacklist(),
            }

    roles = User.Options.roles
    assert len(roles['create']) == 3
    assert len(roles['public']) == 3
    assert len(roles['nospam']) == 2
    assert len(roles['empty']) == 0
    assert len(roles['everything']) == 0

    # Sets sort different with different Python versions. We should be getting something back
    # like: "whitelist('password', 'email', 'name')"
    s = str(roles['create'])
    assert s.startswith('whitelist(') and s.endswith(')')
    assert sorted(s[10:-1].split(', ')) == ["'email'", "'name'", "'password'"]

    # Similar, but now looking for: <Role whitelist('password', 'email', 'name')>
    r = repr(roles['create'])
    assert r.startswith('<Role whitelist(') and r.endswith(')>')
    assert sorted(r[16:-2].split(', ')) == ["'email'", "'name'", "'password'"]

    data = {
        'id': 'NaN',
        'name': 'Arthur',
        'email': 'adent@hitchhiker.gal',
        'password': 'dolphins',
    }

    user = User(
        dict(
            (k, v) for k, v in iteritems(data)
            if k in User._options.roles['create']  # filter by 'create' role
        )
    )

    d = user.serialize(role='public')

    assert d == {
        'id': 42,
        'name': 'Arthur',
        'email': 'adent@hitchhiker.gal',
    }

    d = user.serialize(role='nospam')

    assert d == {
        'id': 42,
        'name': 'Arthur',
    }

    d = user.serialize(role='empty')

    assert d is None

    d = user.serialize(role='everything')

    assert d == {
        'email': 'adent@hitchhiker.gal',
        'id': 42,
        'name': 'Arthur',
        'password': 'dolphins'
    }

    def test_md5_type(self):
        class M(Model):
            md5 = MD5Type()

        import hashlib
        myhash = hashlib.md5("hashthis").hexdigest()
        m = M()
        m.md5 = myhash

        self.assertEqual(m.md5, myhash)
        d = m.serialize()
        self.assertEqual(d, {
            'md5': myhash
        })

        m2 = M(d)
        self.assertEqual(m2.md5, myhash)


def test_serializable_with_list_and_default_role():
    class Player(Model):
        id = LongType()
        display_name = StringType()

        class Options:
            roles = {
                "default": blacklist("id")
            }

    class Game(Model):
        id = StringType()
        result = IntType()
        players = ListType(ModelType(Player))

        class Options:
            roles = {
                "default": blacklist("result")
            }

    p1 = Player({"id": 1, "display_name": "A"})
    p2 = Player({"id": 2, "display_name": "B"})

    game = Game({
        "id": "1",
        "players": [p1, p2]
    })

    assert game.players[0] == p1
    assert game.players[1] == p2

    d = game.serialize(role="default")

    assert d == {
        "id": "1",
        "players": [
            {
                "display_name": "A",
            },
            {
                "display_name": "B",
            },
        ]
    }

    d = game.serialize()

    assert d == {
        "id": "1",
        "players": [
            {
                "display_name": "A",
            },
            {
                "display_name": "B",
            },
        ]
    }



from schematics.exceptions import BaseError, ValidationError


def test_clean_messages():
    err = BaseError(None)

    assert err.clean_messages({'foo': 'bar'}) == {'foo': 'bar'}
    assert err.clean_messages({'foo': ValidationError('bar')}) == {'foo': ['bar']}
    assert err.clean_messages(['foo']) == ['foo']
    assert err.clean_messages([ValidationError('bar')]) == [['bar']]



import pytest

from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.types.compound import ModelType, ListType
from schematics.exceptions import ValidationError, StopValidation


def test_list_field():
    class User(Model):
        ids = ListType(StringType, required=True)

    c = User({
        "ids": []
    })

    c.validate({'ids': []})

    assert c.ids == []


def test_list_with_default_type():
    class CategoryStatsInfo(Model):
        slug = StringType()

    class PlayerInfo(Model):
        categories = ListType(ModelType(CategoryStatsInfo))

    math_stats = CategoryStatsInfo(dict(slug="math"))
    twilight_stats = CategoryStatsInfo(dict(slug="twilight"))
    info = PlayerInfo({
        "categories": [{"slug": "math"}, {"slug": "twilight"}]
    })

    assert info.categories == [math_stats, twilight_stats]

    d = info.serialize()
    assert d == {
        "categories": [{"slug": "math"}, {"slug": "twilight"}],
    }


def test_set_default():
    class CategoryStatsInfo(Model):
        slug = StringType()

    class PlayerInfo(Model):
        categories = ListType(ModelType(CategoryStatsInfo),
                              default=lambda: [],
                              serialize_when_none=True)

    info = PlayerInfo()
    assert info.categories == []

    d = info.serialize()
    assert d == {
        "categories": [],
    }


def test_list_defaults_to_none():
    class PlayerInfo(Model):
        following = ListType(StringType)

    info = PlayerInfo()

    assert info.following is None

    assert info.serialize() == {
        "following": None,
    }


def test_list_default_to_none_embedded_model():
    class QuestionResource(Model):
        url = StringType()

    class QuestionResources(Model):
        pictures = ListType(ModelType(QuestionResource))

    class Question(Model):
        id = StringType()
        resources = ModelType(QuestionResources)

    class QuestionPack(Model):
        id = StringType()
        questions = ListType(ModelType(Question))

    question_pack = QuestionPack({
        "id": "1",
        "questions": [
            {
                "id": "1",
            },
            {
                "id": "2",
                "resources": {
                    "pictures": [],
                }
            },
            {
                "id": "3",
                "resources": {
                    "pictures": [{
                        "url": "http://www.mbl.is/djok",
                    }]
                }
            },
        ]
    })

    assert question_pack.questions[0].resources is None
    assert question_pack.questions[1].resources["pictures"] == []

    resource = QuestionResource({"url": "http://www.mbl.is/djok"})
    assert question_pack.questions[2].resources["pictures"][0] == resource


def test_validation_with_size_limits():
    class User(Model):
        name = StringType()

    class Card(Model):
        users = ListType(ModelType(User), min_size=1, max_size=2, required=True)

    with pytest.raises(ValidationError) as exception:
        c = Card({"users": None})
        c.validate()

    assert exception.value.messages['users'] == [u'This field is required.']

    with pytest.raises(ValidationError) as exception:
        c = Card({"users": []})
        c.validate()

    assert exception.value.messages['users'] == [u'Please provide at least 1 item.']

    with pytest.raises(ValidationError) as exception:
        c = Card({"users": [User(), User(), User()]})
        c.validate()

    assert exception.value.messages['users'] == [u'Please provide no more than 2 items.']


def test_list_field_required():
    class User(Model):
        ids = ListType(StringType(required=True))

    c = User({
        "ids": []
    })

    c.ids = [1]
    c.validate()

    c.ids = [None]
    with pytest.raises(ValidationError):
        c.validate()


def test_list_field_convert():
    class User(Model):
        ids = ListType(IntType)

    c = User({'ids': ["1", "2"]})

    assert c.ids == [1, 2]


def test_list_model_field():
    class User(Model):
        name = StringType()

    class Card(Model):
        users = ListType(ModelType(User), min_size=1, required=True)

    data = {'users': [{'name': u'Doggy'}]}
    c = Card(data)

    c.users = None
    with pytest.raises(ValidationError) as exception:
        c.validate()

    errors = exception.value.messages
    assert errors['users'] == [u'This field is required.']


def test_list_model_field_exception_with_full_message():
    class User(Model):
        name = StringType(max_length=1)

    class Group(Model):
        users = ListType(ModelType(User))

    g = Group({'users': [{'name': "ToLongName"}]})

    with pytest.raises(ValidationError) as exception:
        g.validate()
    assert exception.value.messages == {'users': [{'name': ['String value is too long.']}]}


def test_stop_validation():
    def raiser(x):
        raise StopValidation({'something': 'bad'})

    lst = ListType(StringType(), validators=[raiser])

    with pytest.raises(ValidationError) as exc:
        lst.validate('foo@bar.com')

    assert exc.value.messages == {'something': 'bad'}


def test_compound_fields():
    comments = ListType(ListType, compound_field=StringType)

    assert isinstance(comments.field, ListType)



from schematics.datastructures import OrderedDict
from schematics.transforms import expand, whitelist, flatten
from schematics.models import Model
from schematics.types.serializable import serializable
from schematics.types import StringType, IntType
from schematics.types.compound import (
    ModelType, ListType, EMPTY_LIST, DictType, EMPTY_DICT
)


def test_flattend():
    class Location(Model):
        country_code = StringType()
        region_code = StringType()

    info = Location(dict(country_code="US", region_code="CA"))
    flat_dict = info.flatten()

    assert flat_dict == {
        "country_code": "US",
        "region_code": "CA",
    }

    info_from_flat_dict = Location.from_flat(flat_dict)

    assert type(info) == type(info_from_flat_dict)
    assert info == info_from_flat_dict


def test_flatten_with_root_prefix():
    class Location(Model):
        country_code = StringType()
        region_code = StringType()

    info = Location(dict(country_code="US", region_code="CA"))
    flat_dict = info.flatten(prefix="my_location")

    assert flat_dict == {
        "my_location.country_code": "US",
        "my_location.region_code": "CA",
    }


def test_to_flat_dict_one_level_deep():
    class Location(Model):
        country_code = StringType()
        region_code = StringType()

    class PlayerInfo(Model):
        id = StringType()
        location = ModelType(Location)

    location = Location(dict(country_code="US", region_code="CA"))
    info = PlayerInfo(dict(id=1, location=location))

    flat_dict = info.flatten()

    assert flat_dict == {
        "id": "1",
        "location.country_code": "US",
        "location.region_code": "CA",
    }

    info_from_flat_dict = PlayerInfo.from_flat(flat_dict)

    assert type(info) == type(info_from_flat_dict)
    assert info == info_from_flat_dict
    assert info.location == info_from_flat_dict.location


def test_to_flat_dict_two_level_deep():
    class Something(Model):
        id = IntType()

    class Location(Model):
        country_code = StringType()
        region_code = StringType()

        something = ModelType(Something)

    class PlayerInfo(Model):
        id = StringType()
        location = ModelType(Location)

    location_info = Location(dict(
        country_code="US",
        region_code="CA",
        something={
            "id": 1,
        }
    ))

    info = PlayerInfo(dict(
        id="1",
        location=location_info
    ))

    flat_dict = info.flatten()

    assert flat_dict == {
        "id": "1",
        "location.country_code": "US",
        "location.region_code": "CA",
        "location.something.id": 1,
    }

    info_from_flat_dict = PlayerInfo.from_flat(flat_dict)

    assert type(info) == type(info_from_flat_dict)
    assert info == info_from_flat_dict
    assert info_from_flat_dict.location == location_info


def test_flatten_wiht_listtype_empty_value():
    class PlayerCategoryInfo(Model):
        id = StringType(required=True)
        categories = ListType(IntType, required=True)

    p = PlayerCategoryInfo(dict(id="1", categories=[]))
    flat = p.flatten()

    assert flat == {
        "id": "1",
        "categories": EMPTY_LIST,
    }

    p_from_flat = PlayerCategoryInfo.from_flat(flat)
    assert p_from_flat.categories == []
    assert p == p_from_flat


def test_flatten_wiht_listtype_basic_types():
    class PlayerCategoryInfo(Model):
        id = StringType(required=True)
        categories = ListType(IntType, required=True)

    p = PlayerCategoryInfo(dict(id="1", categories=[1, 2, 3]))
    flat = p.flatten()

    assert flat == {
        "id": "1",
        "categories.0": 1,
        "categories.1": 2,
        "categories.2": 3,
    }

    p_from_flat = PlayerCategoryInfo.from_flat(flat)
    assert p_from_flat.categories == [1, 2, 3]
    assert p == p_from_flat


def test_flatten_with_listtype_after_appending():

    class Game(Model):
        players = ListType(StringType, default=lambda: [])

    game = Game()
    game.players.append("John")

    flat_dict = game.flatten()

    assert flat_dict == {
        "players.0": "John",
    }

    game_from_flat_dict = Game.from_flat(flat_dict)

    assert game == game_from_flat_dict


def test_flatten_with_listtype():
    class ExperienceLevelInfo(Model):
        level = IntType()
        stars = IntType()
        title = StringType()

    class CategoryStatsInfo(Model):
        slug = StringType()

        xp_level = ModelType(ExperienceLevelInfo)

    class PlayerInfo(Model):
        id = StringType()

    class PlayerCategoryInfo(PlayerInfo):
        categories = ListType(ModelType(CategoryStatsInfo))

    input_data = {
        "id": 1,
        "categories": [
        {
            "slug": "math",
            "xp_level": {
                "level": 1,
                "stars": 1,
                "title": "Master",
            }
        },
        {
            "slug": "twilight",
            "xp_level": {
                "level": 2,
                "stars": 1,
                "title": "Master",
            }
        }]
    }
    info = PlayerCategoryInfo(input_data)

    flat_dict = info.flatten()

    assert flat_dict == {
        "id": "1",
        "categories.0.slug": "math",
        "categories.0.xp_level.level": 1,
        "categories.0.xp_level.stars": 1,
        "categories.0.xp_level.title": "Master",
        "categories.1.slug": "twilight",
        "categories.1.xp_level.level": 2,
        "categories.1.xp_level.stars": 1,
        "categories.1.xp_level.title": "Master",
    }

    info_from_flat_dict = PlayerCategoryInfo.from_flat(flat_dict)

    assert info == info_from_flat_dict


def test_flatten_with_dicttype_empty_value():
    class PlayerCategoryInfo(Model):
        id = StringType(required=True)
        categories = DictType(IntType, required=True)

    p = PlayerCategoryInfo(dict(id="1", categories={}))
    flat = p.flatten()

    assert flat == {
        "id": "1",
        "categories": EMPTY_DICT,
    }

    p_from_flat = PlayerCategoryInfo.from_flat(flat)
    assert p_from_flat.categories == {}
    assert p == p_from_flat


def test_flatten_with_dicttype_basic_types():
    class PlayerCategoryInfo(Model):
        id = StringType(required=True)
        categories = DictType(IntType, required=True)

    p = PlayerCategoryInfo(dict(id="1", categories={"a": 1, "b": 2}))
    flat = p.flatten()

    assert flat == {
        "id": "1",
        "categories.a": 1,
        "categories.b": 2
    }

    p_from_flat = PlayerCategoryInfo.from_flat(flat)
    assert p == p_from_flat


def test_flatten_with_dicttype_model_types():
    class CategoryStats(Model):
        total_wins = IntType()

    class PlayerCategoryInfo(Model):
        id = StringType(required=True)
        categories = DictType(ModelType(CategoryStats), required=True)

    p = PlayerCategoryInfo(dict(
        id="1",
        categories={
            "a": {"total_wins": 1},
            "b": {"total_wins": 5},
        }
    ))
    flat = p.flatten()

    assert flat == {
        "id": "1",
        "categories.a.total_wins": 1,
        "categories.b.total_wins": 5,
    }

    p_from_flat = PlayerCategoryInfo.from_flat(flat)
    assert p == p_from_flat


def test_flatten_serializables_on_by_default():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel(dict(level=self.total_points * 2, title="Best"))

    ExperienceLevel(dict(level=4, title="Best"))

    player = Player({"total_points": 2})

    assert player.xp_level.level == 4

    flat = player.flatten()
    assert flat == {"total_points": 2, "xp_level.level": 4, "xp_level.title": "Best"}

    player_from_flat = Player.from_flat(flat)
    assert player == player_from_flat


def test_flatten_ignores_none_by_default():
    class Player(Model):
        id = StringType()
        display_name = StringType()

    player = Player({"display_name": "Joe"})

    assert player.id is None

    flat = player.flatten()
    assert flat == {"display_name": "Joe"}

    player_from_flat = Player.from_flat(flat)
    assert player, player_from_flat


def test_flatten_includes_none_when_asked():
    class Player(Model):
        id = StringType()
        display_name = StringType()

    player = Player({"display_name": "Joe"})

    flat = flatten(Player, player, ignore_none=False)

    assert flat == {'display_name': 'Joe', 'id': None}


def test_flatten_with_whitelist():
    class TopicStats(Model):
        total_points = IntType()
        total_wins = IntType()
        total_losses = IntType()

        class Options:
            roles = {
                "public": whitelist("total_wins", "total_losses"),
            }

        @serializable
        def games_played(self):
            return self.total_wins + self.total_losses

    topic_stats = TopicStats(dict(
        total_points=2,
        total_wins=3,
        total_losses=4,
    ))

    flat = topic_stats.flatten(role="public")
    assert flat == {
        "total_wins": 3,
        "total_losses": 4,
    }


def test_flatten_dicts_coercers_keys_to_strings():

    class Player(Model):
        id = StringType()

    class Game(Model):
        id = StringType()
        players = DictType(ModelType(Player), coerce_key=lambda k: int(k))

    g = Game(dict(
        id="1",
        players={
            1: {
                "id": 1,
            }
        }
    ))

    assert g.id == "1"
    assert g.players == {1: Player(dict(id="1"))}

    flat = g.flatten()

    assert flat == {
        "id": "1",
        "players.1.id": "1",
    }

    g_from_flat = Game.from_flat(flat)
    assert g == g_from_flat


def test_expand_with_both_empty_dict_and_values():
    different_orderings = [
        [
            ("categories", '{}'),
            ("categories.basketball.category_slug", 'basketball'),
            ("categories.basketball.total_draws", '0'),
            ("categories.basketball.total_losses", '2'),
        ],
        [
            ("categories.basketball.category_slug", 'basketball'),
            ("categories", '{}'),
            ("categories.basketball.total_draws", '0'),
            ("categories.basketball.total_losses", '2'),
        ],
        [
            ("categories.basketball.category_slug", 'basketball'),
            ("categories.basketball.total_draws", '0'),
            ("categories.basketball.total_losses", '2'),
            ("categories", '{}'),
        ]
    ]

    for ordering in different_orderings:
        flat_data = OrderedDict(ordering)

        expanded = expand(flat_data)
        assert expanded == {
            "categories": {
                "basketball": {
                    "category_slug": "basketball",
                    "total_draws": "0",
                    "total_losses": "2",
                }
            }
        }


def test_expand_with_both_empty_list_and_values():
    different_orderings = [
        [
            ("categories", '[]'),
            ("categories.0", 'basketball'),
            ("categories.1", '0'),
            ("categories.2", '2'),
        ],
        [
            ("categories.0", 'basketball'),
            ("categories", '[]'),
            ("categories.1", '0'),
            ("categories.2", '2'),
        ],
        [
            ("categories.0", 'basketball'),
            ("categories.1", '0'),
            ("categories.2", '2'),
            ("categories", '[]'),
        ]
    ]
    for ordering in different_orderings:
        flat_data = OrderedDict(ordering)

        expanded = expand(flat_data)
        assert expanded == {
            "categories": {
                "0": "basketball",
                "1": "0",
                "2": "2",
            }
        }



from six import iteritems

class BaseError(Exception):

    def __init__(self, messages):
        if not isinstance(messages, (list, tuple, dict)):
            messages = [messages]

        clean_messages = self.clean_messages(messages)

        Exception.__init__(self, clean_messages)
        self.messages = clean_messages

    def clean_messages(self, messages):
        if isinstance(messages, dict):
            clean_messages = {}
            for key, value in iteritems(messages):
                if isinstance(value, ValidationError):
                    value = value.messages
                clean_messages[key] = value
        else:
            clean_messages = []
            for message in messages:
                if isinstance(message, ValidationError):
                    message = message.messages
                clean_messages.append(message)

        return clean_messages


class ConversionError(BaseError, TypeError):

    """ Exception raised when data cannot be converted to the correct python type """
    pass


class ModelConversionError(ConversionError):
    def __init__(self, messages, partial_data=None):
        super(ModelConversionError, self).__init__(messages)
        self.partial_data = partial_data


class ValidationError(BaseError, ValueError):

    """Exception raised when invalid data is encountered."""
    pass


class ModelValidationError(ValidationError):
    pass


class StopValidation(ValidationError):

    """Exception raised when no more validation need occur."""
    pass


class MockCreationError(ValueError):
    """Exception raised when a mock value cannot be generated."""
    pass



import uuid
import re
import datetime
import decimal
import itertools
import functools
import random
import string

import six
from six import iteritems

from ..exceptions import (
    StopValidation, ValidationError, ConversionError, MockCreationError
)

try:
    from string import ascii_letters # PY3
except ImportError:
    from string import letters as ascii_letters #PY2

try:
    basestring #PY2
except NameError:
    basestring = str #PY3

try:
    unicode #PY2
except:
    import codecs
    unicode = str #PY3

def utf8_decode(s):

    if six.PY3:
        s = str(s) #todo: right thing to do?
    else:
        s = unicode(s, 'utf-8')

    return s

def fill_template(template, min_length, max_length):
    return template % random_string(
        get_value_in(
            min_length,
            max_length,
            padding=len(template) - 2,
            required_length=1))


def force_unicode(obj, encoding='utf-8'):
    if isinstance(obj, basestring):
        if not isinstance(obj, unicode):
            #obj = unicode(obj, encoding)
            obj = utf8_decode(obj)
    elif not obj is None:
        #obj = unicode(obj)
        obj = utf8_decode(obj)

    return obj


def get_range_endpoints(min_length, max_length, padding=0, required_length=0):
    if min_length is None and max_length is None:
        min_length = 0
        max_length = 16
    elif min_length is None:
        min_length = 0
    elif max_length is None:
        max_length = max(min_length * 2, 16)

    if padding:
        max_length = max_length - padding
        min_length = max(min_length - padding, 0)

    if max_length < required_length:
        raise MockCreationError(
            'This field is too short to hold the mock data')

    min_length = max(min_length, required_length)

    return min_length, max_length


def get_value_in(min_length, max_length, padding=0, required_length=0):
    return random.randint(
        *get_range_endpoints(min_length, max_length, padding, required_length))


def random_string(length, chars=ascii_letters + string.digits):
    return ''.join(random.choice(chars) for _ in range(length))


_last_position_hint = -1
_next_position_hint = itertools.count()


class TypeMeta(type):

    """
    Meta class for BaseType. Merges `MESSAGES` dict and accumulates
    validator methods.
    """

    def __new__(mcs, name, bases, attrs):
        messages = {}
        validators = []

        for base in reversed(bases):
            if hasattr(base, 'MESSAGES'):
                messages.update(base.MESSAGES)

            if hasattr(base, "_validators"):
                validators.extend(base._validators)

        if 'MESSAGES' in attrs:
            messages.update(attrs['MESSAGES'])

        attrs['MESSAGES'] = messages

        for attr_name, attr in iteritems(attrs):
            if attr_name.startswith("validate_"):
                validators.append(attr)

        attrs["_validators"] = validators

        return type.__new__(mcs, name, bases, attrs)


class BaseType(TypeMeta('BaseTypeBase', (object, ), {})):

    """A base class for Types in a Schematics model. Instances of this
    class may be added to subclasses of ``Model`` to define a model schema.

    Validators that need to access variables on the instance
    can be defined be implementing methods whose names start with ``validate_``
    and accept one parameter (in addition to ``self``)

    :param required:
        Invalidate field when value is None or is not supplied. Default:
        False.
    :param default:
        When no data is provided default to this value. May be a callable.
        Default: None.
    :param serialized_name:
        The name of this field defaults to the class attribute used in the
        model. However if the field has another name in foreign data set this
        argument. Serialized data will use this value for the key name too.
    :param deserialize_from:
        A name or list of named fields for which foreign data sets are
        searched to provide a value for the given field.  This only effects
        inbound data.
    :param choices:
        A list of valid choices. This is the last step of the validator
        chain.
    :param validators:
        A list of callables. Each callable receives the value after it has been
        converted into a rich python type. Default: []
    :param serialize_when_none:
        Dictates if the field should appear in the serialized data even if the
        value is None. Default: True
    :param messages:
        Override the error messages with a dict. You can also do this by
        subclassing the Type and defining a `MESSAGES` dict attribute on the
        class. A metaclass will merge all the `MESSAGES` and override the
        resulting dict with instance level `messages` and assign to
        `self.messages`.

    """

    MESSAGES = {
        'required': u"This field is required.",
        'choices': u"Value must be one of {0}.",
    }

    def __init__(self, required=False, default=None, serialized_name=None,
                 choices=None, validators=None, deserialize_from=None,
                 serialize_when_none=None, messages=None):
        super(BaseType, self).__init__()
        self.required = required
        self._default = default
        self.serialized_name = serialized_name
        if choices and not isinstance(choices, (list, tuple)):
            raise TypeError('"choices" must be a list or tuple')
        self.choices = choices
        self.deserialize_from = deserialize_from

        self.validators = [functools.partial(v, self) for v in self._validators]
        if validators:
            self.validators += validators

        self.serialize_when_none = serialize_when_none
        self.messages = dict(self.MESSAGES, **(messages or {}))
        self._position_hint = next(_next_position_hint)  # For ordering of fields

    def __call__(self, value):
        return self.to_native(value)

    def _mock(self, context=None):
        return None

    @property
    def default(self):
        default = self._default
        if callable(self._default):
            default = self._default()
        return default

    def to_primitive(self, value, context=None):
        """Convert internal data to a value safe to serialize.
        """
        return value

    def to_native(self, value, context=None):
        """
        Convert untrusted data to a richer Python construct.
        """
        return value

    def allow_none(self):
        if hasattr(self, 'owner_model'):
            return self.owner_model.allow_none(self)
        else:
            return self.serialize_when_none

    def validate(self, value):
        """
        Validate the field and return a clean value or raise a
        ``ValidationError`` with a list of errors raised by the validation
        chain. Stop the validation process from continuing through the
        validators by raising ``StopValidation`` instead of ``ValidationError``.

        """

        errors = []

        for validator in self.validators:
            try:
                validator(value)
            except ValidationError as exc:
                errors.extend(exc.messages)

                if isinstance(exc, StopValidation):
                    break

        if errors:
            raise ValidationError(errors)

    def validate_required(self, value):
        if self.required and value is None:
            raise ValidationError(self.messages['required'])

    def validate_choices(self, value):
        if self.choices is not None:
            if value not in self.choices:
                raise ValidationError(self.messages['choices']
                                      .format(unicode(self.choices)))

    def mock(self, context=None):
        if not self.required and not random.choice([True, False]):
            return self.default
        if self.choices is not None:
            return random.choice(self.choices)
        return self._mock(context)


class UUIDType(BaseType):

    """A field that stores a valid UUID value.
    """
    MESSAGES = {
        'convert': u"Couldn't interpret '{0}' value as UUID.",
    }

    def _mock(self, context=None):
        return uuid.uuid4()

    def to_native(self, value, context=None):
        if not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value)
            except (AttributeError, TypeError, ValueError):
                raise ConversionError(self.messages['convert'].format(value))
        return value

    def to_primitive(self, value, context=None):
        return str(value)


class IPv4Type(BaseType):

    """ A field that stores a valid IPv4 address """

    def _mock(self, context=None):
        return '.'.join(str(random.randrange(256)) for _ in range(4))

    @classmethod
    def valid_ip(cls, addr):
        try:
            addr = addr.strip().split(".")
        except AttributeError:
            return False
        try:
            return len(addr) == 4 and all(0 <= int(octet) < 256 for octet in addr)
        except ValueError:
            return False

    def validate(self, value):
        """
          Make sure the value is a IPv4 address:
          http://stackoverflow.com/questions/9948833/validate-ip-address-from-list
        """
        if not IPv4Type.valid_ip(value):
            error_msg = 'Invalid IPv4 address'
            raise ValidationError(error_msg)
        return True


class StringType(BaseType):

    """A unicode string field. Default minimum length is one. If you want to
    accept empty strings, init with ``min_length`` 0.
    """

    allow_casts = (int, str)

    MESSAGES = {
        'convert': u"Couldn't interpret '{0}' as string.",
        'max_length': u"String value is too long.",
        'min_length': u"String value is too short.",
        'regex': u"String value did not match validation regex.",
    }

    def __init__(self, regex=None, max_length=None, min_length=None, **kwargs):
        self.regex = regex
        self.max_length = max_length
        self.min_length = min_length

        super(StringType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return random_string(get_value_in(self.min_length, self.max_length))

    def to_native(self, value, context=None):
        if value is None:
            return None

        if not isinstance(value, unicode):
            if isinstance(value, self.allow_casts):
                if not isinstance(value, str):
                    value = str(value)
                value = utf8_decode(value) #unicode(value, 'utf-8')
            else:
                raise ConversionError(self.messages['convert'].format(value))

        return value

    def validate_length(self, value):
        len_of_value = len(value) if value else 0

        if self.max_length is not None and len_of_value > self.max_length:
            raise ValidationError(self.messages['max_length'])

        if self.min_length is not None and len_of_value < self.min_length:
            raise ValidationError(self.messages['min_length'])

    def validate_regex(self, value):
        if self.regex is not None and re.match(self.regex, value) is None:
            raise ValidationError(self.messages['regex'])


class URLType(StringType):

    """A field that validates input as an URL.

    If verify_exists=True is passed the validate function will make sure
    the URL makes a valid connection.
    """

    MESSAGES = {
        'invalid_url': u"Not a well formed URL.",
        'not_found': u"URL does not exist.",
    }

    URL_REGEX = re.compile(
        r'^https?://'
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,2000}[A-Z0-9])?\.)+[A-Z]{2,63}\.?|'
        r'localhost|'
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
        r'(?::\d+)?'
        r'(?:/?|[/?]\S+)$', re.IGNORECASE
    )

    def __init__(self, verify_exists=False, **kwargs):
        self.verify_exists = verify_exists
        super(URLType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return fill_template('http://a%s.ZZ', self.min_length,
                             self.max_length)

    def validate_url(self, value):
        if not URLType.URL_REGEX.match(value):
            raise StopValidation(self.messages['invalid_url'])
        if self.verify_exists:
            from six.moves import urllib
            try:
                request = urllib.Request(value)
                urllib.urlopen(request)
            except Exception:
                raise StopValidation(self.messages['not_found'])


class EmailType(StringType):

    """A field that validates input as an E-Mail-Address.
    """

    MESSAGES = {
        'email': u"Not a well formed email address."
    }

    EMAIL_REGEX = re.compile(
        # dot-atom
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"
        # quoted-string
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016'
        r'-\177])*"'
        # domain
        r')@(?:[A-Z0-9](?:[A-Z0-9-]{0,2000}[A-Z0-9])?\.)+[A-Z]{2,63}\.?$',
        re.IGNORECASE
    )

    def _mock(self, context=None):
        return fill_template('%s@example.com', self.min_length,
                             self.max_length)

    def validate_email(self, value):
        if not EmailType.EMAIL_REGEX.match(value):
            raise StopValidation(self.messages['email'])


class NumberType(BaseType):

    """A number field.
    """

    MESSAGES = {
        'number_coerce': u"Value '{0}' is not {1}.",
        'number_min': u"{0} value should be greater than {1}.",
        'number_max': u"{0} value should be less than {1}.",
    }

    def __init__(self, number_class, number_type,
                 min_value=None, max_value=None, **kwargs):
        self.number_class = number_class
        self.number_type = number_type
        self.min_value = min_value
        self.max_value = max_value

        super(NumberType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return get_value_in(self.min_value, self.max_value)

    def to_native(self, value, context=None):
        try:
            value = self.number_class(value)
        except (TypeError, ValueError):
            raise ConversionError(self.messages['number_coerce']
                                  .format(value, self.number_type.lower()))

        return value

    def validate_is_a_number(self, value):
        try:
            self.number_class(value)
        except (TypeError, ValueError):
            raise ConversionError(self.messages['number_coerce']
                                  .format(value, self.number_type.lower()))

    def validate_range(self, value):
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(self.messages['number_min']
                                  .format(self.number_type, self.min_value))

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(self.messages['number_max']
                                  .format(self.number_type, self.max_value))

        return value


class IntType(NumberType):

    """A field that validates input as an Integer
    """

    def __init__(self, *args, **kwargs):
        super(IntType, self).__init__(number_class=int,
                                      number_type='Int',
                                      *args, **kwargs)


class LongType(NumberType):

    """A field that validates input as a Long
    """

    def __init__(self, *args, **kwargs):

        try:
            number_class = long #PY2
        except NameError:
            number_class = int #PY3


        super(LongType, self).__init__(number_class=number_class,
                                       number_type='Long',
                                       *args, **kwargs)


class FloatType(NumberType):

    """A field that validates input as a Float
    """

    def __init__(self, *args, **kwargs):
        super(FloatType, self).__init__(number_class=float,
                                        number_type='Float',
                                        *args, **kwargs)


class DecimalType(BaseType):

    """A fixed-point decimal number field.
    """

    MESSAGES = {
        'number_coerce': u"Number '{0}' failed to convert to a decimal.",
        'number_min': u"Value should be greater than {0}.",
        'number_max': u"Value should be less than {0}.",
    }

    def __init__(self, min_value=None, max_value=None, **kwargs):
        self.min_value, self.max_value = min_value, max_value

        super(DecimalType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return get_value_in(self.min_value, self.max_value)

    def to_primitive(self, value, context=None):
        return unicode(value)

    def to_native(self, value, context=None):
        if not isinstance(value, decimal.Decimal):
            if not isinstance(value, basestring):
                value = unicode(value)
            try:
                value = decimal.Decimal(value)
            except (TypeError, decimal.InvalidOperation):
                raise ConversionError(self.messages['number_coerce'].format(value))

        return value

    def validate_range(self, value):
        if self.min_value is not None and value < self.min_value:
            error_msg = self.messages['number_min'].format(self.min_value)
            raise ValidationError(error_msg)

        if self.max_value is not None and value > self.max_value:
            error_msg = self.messages['number_max'].format(self.max_value)
            raise ValidationError(error_msg)

        return value


class HashType(BaseType):

    MESSAGES = {
        'hash_length': u"Hash value is wrong length.",
        'hash_hex': u"Hash value is not hexadecimal.",
    }

    def _mock(self, context=None):
        return random_string(self.LENGTH, string.hexdigits)

    def to_native(self, value, context=None):
        if len(value) != self.LENGTH:
            raise ValidationError(self.messages['hash_length'])
        try:
            int(value, 16)
        except ValueError:
            raise ConversionError(self.messages['hash_hex'])
        return value


class MD5Type(HashType):

    """A field that validates input as resembling an MD5 hash.
    """

    LENGTH = 32


class SHA1Type(HashType):

    """A field that validates input as resembling an SHA1 hash.
    """

    LENGTH = 40


class BooleanType(BaseType):

    """A boolean field type. In addition to ``True`` and ``False``, coerces these
    values:

    + For ``True``: "True", "true", "1"
    + For ``False``: "False", "false", "0"

    """

    TRUE_VALUES = ('True', 'true', '1')
    FALSE_VALUES = ('False', 'false', '0')

    def _mock(self, context=None):
        return random.choice([True, False])

    def to_native(self, value, context=None):
        if isinstance(value, basestring):
            if value in self.TRUE_VALUES:
                value = True
            elif value in self.FALSE_VALUES:
                value = False

        if isinstance(value, int) and value in [0, 1]:
            value = bool(value)

        if not isinstance(value, bool):
            raise ConversionError(u"Must be either true or false.")

        return value


class DateType(BaseType):

    """Defaults to converting to and from ISO8601 date values.
    """

    SERIALIZED_FORMAT = '%Y-%m-%d'
    MESSAGES = {
        'parse': u"Could not parse {0}. Should be ISO8601 (YYYY-MM-DD).",
    }

    def __init__(self, **kwargs):
        self.serialized_format = self.SERIALIZED_FORMAT
        super(DateType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return datetime.datetime(
            year=random.randrange(600) + 1900,
            month=random.randrange(12) + 1,
            day=random.randrange(28) + 1,
        )

    def to_native(self, value, context=None):
        if isinstance(value, datetime.date):
            return value

        try:
            return datetime.datetime.strptime(value, self.serialized_format).date()
        except (ValueError, TypeError):
            raise ConversionError(self.messages['parse'].format(value))

    def to_primitive(self, value, context=None):
        return value.strftime(self.serialized_format)


class DateTimeType(BaseType):

    """Defaults to converting to and from ISO8601 datetime values.

    :param formats:
        A value or list of values suitable for ``datetime.datetime.strptime``
        parsing. Default: `('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%dT%H:%M:%S',
        '%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%SZ')`
    :param serialized_format:
        The output format suitable for Python ``strftime``. Default: ``'%Y-%m-%dT%H:%M:%S.%f'``

    """

    DEFAULT_FORMATS = (
        '%Y-%m-%dT%H:%M:%S.%f',  '%Y-%m-%dT%H:%M:%S',
        '%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%SZ',
    )
    SERIALIZED_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

    MESSAGES = {
        'parse_formats': u'Could not parse {0}. Valid formats: {1}',
        'parse': u"Could not parse {0}. Should be ISO8601.",
    }

    def __init__(self, formats=None, serialized_format=None, **kwargs):
        """

        """
        if isinstance(formats, basestring):
            formats = [formats]
        if formats is None:
            formats = self.DEFAULT_FORMATS
        if serialized_format is None:
            serialized_format = self.SERIALIZED_FORMAT
        self.formats = formats
        self.serialized_format = serialized_format
        super(DateTimeType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return datetime.datetime(
            year=random.randrange(600) + 1900,
            month=random.randrange(12) + 1,
            day=random.randrange(28) + 1,
            hour=random.randrange(24),
            minute=random.randrange(60),
            second=random.randrange(60),
            microsecond=random.randrange(1000000),
        )

    def to_native(self, value, context=None):
        if isinstance(value, datetime.datetime):
            return value

        for fmt in self.formats:
            try:
                return datetime.datetime.strptime(value, fmt)
            except (ValueError, TypeError):
                continue
        if self.formats == self.DEFAULT_FORMATS:
            message = self.messages['parse'].format(value)
        else:
            message = self.messages['parse_formats'].format(
                value, ", ".join(self.formats)
            )
        raise ConversionError(message)

    def to_primitive(self, value, context=None):
        if callable(self.serialized_format):
            return self.serialized_format(value)
        return value.strftime(self.serialized_format)


class GeoPointType(BaseType):

    """A list storing a latitude and longitude.
    """

    def _mock(self, context=None):
        return (random.randrange(-90, 90), random.randrange(-90, 90))

    def to_native(self, value, context=None):
        """Make sure that a geo-value is of type (x, y)
        """
        if not len(value) == 2:
            raise ValueError('Value must be a two-dimensional point')
        if isinstance(value, dict):
            for val in value.values():
                if not isinstance(val, (float, int)):
                    raise ValueError('Both values in point must be float or int')
        elif isinstance(value, (list, tuple)):
            if (not isinstance(value[0], (float, int)) or
                    not isinstance(value[1], (float, int))):
                raise ValueError('Both values in point must be float or int')
        else:
            raise ValueError('GeoPointType can only accept tuples, lists, or dicts')

        return value


class MultilingualStringType(BaseType):

    """
    A multilanguage string field, stored as a dict with {'locale': 'localized_value'}.

    Minimum and maximum lengths apply to each of the localized values.

    At least one of ``default_locale`` or ``context['locale']`` must be defined
    when calling ``.to_primitive``.

    """

    allow_casts = (int, str)

    MESSAGES = {
        'convert': u"Couldn't interpret value as string.",
        'max_length': u"String value in locale {0} is too long.",
        'min_length': u"String value in locale {0} is too short.",
        'locale_not_found': u"No requested locale was available.",
        'no_locale': u"No default or explicit locales were given.",
        'regex_locale': u"Name of locale {0} did not match validation regex.",
        'regex_localized': u"String value in locale {0} did not match validation regex.",
    }

    LOCALE_REGEX = r'^[a-z]{2}(:?_[A-Z]{2})?$'

    def __init__(self, regex=None, max_length=None, min_length=None,
                 default_locale=None, locale_regex=LOCALE_REGEX, **kwargs):
        self.regex = re.compile(regex) if regex else None
        self.max_length = max_length
        self.min_length = min_length
        self.default_locale = default_locale
        self.locale_regex = re.compile(locale_regex) if locale_regex else None

        super(MultilingualStringType, self).__init__(**kwargs)

    def _mock(self, context=None):
        return random_string(get_value_in(self.min_length, self.max_length))

    def to_native(self, value, context=None):
        """Make sure a MultilingualStringType value is a dict or None."""

        if not (value is None or isinstance(value, dict)):
            raise ValueError('Value must be a dict or None')

        return value

    def to_primitive(self, value, context=None):
        """
        Use a combination of ``default_locale`` and ``context['locale']`` to return
        the best localized string.

        """
        if value is None:
            return None

        context_locale = None
        if context is not None and 'locale' in context:
            context_locale = context['locale']

        # Build a list of all possible locales to try
        possible_locales = []
        for locale in (context_locale, self.default_locale):
            if not locale:
                continue

            if isinstance(locale, basestring):
                possible_locales.append(locale)
            else:
                possible_locales.extend(locale)

        if not possible_locales:
            raise ConversionError(self.messages['no_locale'])

        for locale in possible_locales:
            if locale in value:
                localized = value[locale]
                break
        else:
            raise ConversionError(self.messages['locale_not_found'])

        if not isinstance(localized, unicode):
            if isinstance(localized, self.allow_casts):
                if not isinstance(localized, str):
                    localized = str(localized)
                #localized = unicode(localized, 'utf-8')
                localized = utf8_decode(localized)
            else:
                raise ConversionError(self.messages['convert'])

        return localized

    def validate_length(self, value):
        for locale, localized in value.items():
            len_of_value = len(localized) if localized else 0

            if self.max_length is not None and len_of_value > self.max_length:
                raise ValidationError(self.messages['max_length'].format(locale))

            if self.min_length is not None and len_of_value < self.min_length:
                raise ValidationError(self.messages['min_length'].format(locale))

    def validate_regex(self, value):
        if self.regex is None and self.locale_regex is None:
            return

        for locale, localized in value.items():
            if self.regex is not None and self.regex.match(localized) is None:
                raise ValidationError(
                    self.messages['regex_localized'].format(locale))

            if self.locale_regex is not None and self.locale_regex.match(locale) is None:
                raise ValidationError(
                    self.messages['regex_locale'].format(locale))



import pickle
from copy import deepcopy
from six import PY3
import pytest

from schematics.datastructures import OrderedDict


def test_od_create():
    od = OrderedDict()
    assert od.keys() == []

    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    assert od.keys() == ['a', 'c', 'foo']

    od = OrderedDict(a='b', c='d', foo='bar')
    assert set(od.keys()) == set(['a', 'c', 'foo'])


def test_od_delete_key():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    del od['c']
    assert od.keys() == ['a', 'foo']


def test_od_deepcopy():
    lst = [1, 2, 3]
    od = OrderedDict(a=lst)
    new_od = deepcopy(od)
    new_od['a'].append(4)

    assert od['a'] is lst
    assert lst == [1, 2, 3]
    assert new_od['a'] == [1, 2, 3, 4]
    assert od.keys() == new_od.keys()


def test_od_reduce():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    serialized = pickle.dumps(od)
    new_od = pickle.loads(serialized)

    assert od == new_od
    assert isinstance(new_od, OrderedDict)


def test_od_reversed():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])

    assert list(reversed(od)) == ['foo', 'c', 'a']


def test_od_fromkeys():
    od = OrderedDict.fromkeys(['a', 'b', 'c'], 'foo')

    assert od == {'a': 'foo', 'b': 'foo', 'c': 'foo'}


def test_od_clear():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    od.clear()

    assert od == {}


def test_od_copy():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', object())])
    new_od = od.copy()

    assert od == new_od
    assert od['foo'] is new_od['foo']


def test_od_pop():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])

    assert od.pop('a') == 'b'
    assert 'a' not in od
    assert od.keys() == ['c', 'foo']

    assert od.pop('bar', 1) == 1

    with pytest.raises(KeyError):
        od.pop('bar')


def test_od_popitem():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])

    assert od.popitem() == ('a', 'b')
    assert 'a' not in od
    assert od.keys() == ['c', 'foo']

    assert od.popitem() == ('c', 'd')
    assert 'c' not in od
    assert od.keys() == ['foo']

    assert od.popitem() == ('foo', 'bar')
    assert 'foo' not in od
    assert od.keys() == []

    with pytest.raises(KeyError):
        od.popitem()


def test_od_setdefault():
    od = OrderedDict([('a', 'b')])

    assert od.setdefault('foo') is None
    assert od.keys() == ['a', 'foo']

    assert od.setdefault('bar', 'baz') == 'baz'
    assert od.keys() == ['a', 'foo', 'bar']

    assert od.setdefault('a') == 'b'
    assert od == {'a': 'b', 'foo': None, 'bar': 'baz'}


def test_od_update():
    od = OrderedDict()
    with pytest.raises(TypeError):
        od.update([], [])


def test_sort():
    items = []
    for i in range(10):
        items.append((i, 9 - i))

    od = OrderedDict(items)
    od.sort(key=lambda x: x[1])

    assert od.keys() == list(range(9, -1, -1))

    items = []
    for i in range(10):
        items.append((9 - i, 9 - i))

    od = OrderedDict(items)
    od.sort()

    assert od.keys() == list(range(10))

    od = OrderedDict(items)
    od.sort(reverse=True)

    assert od.keys() == list(range(9, -1, -1))


def test_repr():
    od = OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])

    assert repr(od) == "OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])"



import pytest

from schematics.types.compound import MultiType


def test_base_does_not_implement_export_loop():
    with pytest.raises(NotImplementedError):
        MultiType().export_loop(None, None)



import pytest

from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.types.compound import ModelType, ListType
from schematics.exceptions import ValidationError


def test_simple_embedded_models():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    r = repr(Player.location)
    assert r.startswith("<schematics.types.compound.ModelType object at 0x")
    #assert r.endswith("for <class 'tests.test_model_type.Location'>>") #this assertion not compatible with PY3

    p = Player(dict(id=1, location={"country_code": "US"}))

    assert p.id == 1
    assert p.location.country_code == "US"

    p.location = Location({"country_code": "IS"})

    assert isinstance(p.location, Location)
    assert p.location.country_code == "IS"

    assert Player.location.to_native(None) is None


def test_simple_embedded_models_is_none():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p = Player(dict(id=1))

    assert p.id == 1
    assert p.location is None


def test_simple_embedded_model_set_to_none():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p = Player(dict(id=1))
    p.location = None

    assert p.id == 1
    assert p.location is None


def test_simple_embedded_model_is_none_within_listtype():
    class QuestionResources(Model):
        type = StringType()

    class Question(Model):
        id = StringType()
        resources = ModelType(QuestionResources)

    class QuestionPack(Model):
        id = StringType()
        questions = ListType(ModelType(Question))

    question_pack = QuestionPack({
        "id": "1",
        "questions": [
            {
                "id": "1",
            },
        ]
    })

    assert question_pack.questions[0].resources is None


def test_raises_validation_error_on_init_with_partial_submodel():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)

    class Card(Model):
        user = ModelType(User)

    u = User({'name': 'Arthur'})
    c = Card({'user': u})

    with pytest.raises(ValidationError):
        c.validate()


def test_model_type():
    class User(Model):
        name = StringType()

    class Card(Model):
        user = ModelType(User)

    c = Card({"user": {'name': u'Doggy'}})
    assert isinstance(c.user, User)
    assert c.user.name == "Doggy"


def test_equality_with_embedded_models():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p1 = Player(dict(id=1, location={"country_code": "US"}))
    p2 = Player(dict(id=1, location={"country_code": "US"}))

    assert id(p1.location) != id(p2.location)
    assert p1.location == p2.location

    assert p1 == p2


def test_default_value_when_embedded_model():
    class Question(Model):
        question_id = StringType(required=True)

        type = StringType(default="text")

    class QuestionPack(Model):

        question = ModelType(Question)

    pack = QuestionPack({
        "question": {
            "question_id": 1
        }
    })

    assert pack.question.question_id == "1"
    assert pack.question.type == "text"


def test_export_loop_with_subclassed_model():
    class Asset(Model):
        file_name = StringType()

    class S3Asset(Asset):
        bucket_name = StringType()

    class Product(Model):
        title = StringType()
        asset = ModelType(Asset)

    asset = S3Asset({'bucket_name': 'assets_bucket', 'file_name': 'bar'})

    product = Product({'title': 'baz', 'asset': asset})

    primitive = product.to_primitive()
    assert 'bucket_name' in primitive['asset']

    native = product.to_native()
    assert 'bucket_name' in native['asset']



from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.validate import validate
from schematics.exceptions import ValidationError


def test_validate_simple_dict():
    class Player(Model):
        id = IntType()

    validate(Player, {'id': 4})


def test_validate_keep_context_data():
    class Player(Model):
        id = IntType()
        name = StringType()

    p1 = Player({'id': 4})
    data = validate(Player, {'name': 'Arthur'}, context=p1._data)

    assert data == {'id': 4, 'name': 'Arthur'}
    assert data != p1._data


def test_validate_override_context_data():
    class Player(Model):
        id = IntType()

    p1 = Player({'id': 4})
    data = validate(Player, {'id': 3}, context=p1._data)

    assert data == {'id': 3}


def test_validate_ignore_extra_context_data():
    class Player(Model):
        id = IntType()

    data = validate(Player, {'id': 4}, context={'name': 'Arthur'})

    assert data == {'id': 4, 'name': 'Arthur'}


def test_validate_strict_with_context_data():
    class Player(Model):
        id = IntType()

    try:
        validate(Player, {'id': 4}, strict=True, context={'name': 'Arthur'})
    except ValidationError as e:
        assert 'name' in e.messages


def test_validate_partial_with_context_data():
    class Player(Model):
        id = IntType()
        name = StringType(required=True)

    data = validate(Player, {'id': 4}, partial=False, context={'name': 'Arthur'})

    assert data == {'id': 4, 'name': 'Arthur'}


def test_validate_with_instance_level_validators():
    class Player(Model):
        id = IntType()

        def validate_id(self, context, value):
            if p1._initial['id'] != value:
                p1._data['id'] = p1._initial['id']
                raise ValidationError('Cannot change id')

    p1 = Player({'id': 4})
    p1.id = 3

    try:
        validate(Player, p1)
    except ValidationError as e:
        assert 'id' in e.messages
        assert 'Cannot change id' in e.messages['id']
        assert p1.id == 4



# -*- coding: utf-8 -*-
import pytest

from schematics.models import Model
from schematics.types.base import StringType
from schematics.types.compound import ListType, ModelType
from schematics.types.serializable import serializable
from schematics.exceptions import ValidationError


def test_reason_why_we_must_bind_fields():
    class Person(Model):
        name = StringType(required=True)

    p1 = Person()
    p2 = Person()

    assert p1 == p2
    assert id(p1) != id(p2)

    assert p1.name == p2.name
    assert id(p1.name) == id(p2.name)

    p1.name = "Jóhann"
    p1.validate()
    with pytest.raises(ValidationError):
        p2.validate()

    assert p1 != p2
    assert id(p1) != id(p2)
    assert p1.name != p2.name
    assert id(p1.name) != id(p2.name)

    assert id(p1._fields["name"]) == id(p2._fields["name"])


def test_reason_why_we_must_bind_fields_model_field():
    class Location(Model):
        country_code = StringType(required=True)

    class Person(Model):
        location = ModelType(Location)

    p1 = Person()
    p2 = Person()

    assert p1 == p2
    assert id(p1) != id(p2)

    assert p1.location == p2.location
    assert id(p1.location) == id(p2.location)

    p1.location = {"country_code": "us"}
    assert p1.location.country_code == "us"

    p2.location = {}

    with pytest.raises(ValidationError):
        p2.validate()

    assert p1 != p2
    assert id(p1) != id(p2)
    assert p1.location != p2.location
    assert id(p1.location) != id(p2.location)

    assert id(p1._fields["location"]) == id(p2._fields["location"])


def test_field_binding():
    class Person(Model):
        name = StringType(required=True)

    class Course(Model):
        id = StringType(required=True, validators=[])
        attending = ListType(ModelType(Person))

    class School(Model):
        courses = ListType(ModelType(Course))

    valid_data = {
        'courses': [
            {'id': 'ENG103', 'attending': [
                {'name': u'Danny'},
                {'name': u'Sandy'}]},
            {'id': 'ENG203', 'attending': [
                {'name': u'Danny'},
                {'name': u'Sandy'}
            ]}
        ]
    }

    new_school = School(valid_data)
    new_school.validate()

    school = School(valid_data)
    school.validate()

    assert id(new_school) != id(school)

    assert id(new_school.courses[0].attending[0]) != id(school.courses[0].attending[0])


def test_serializable_doesnt_keep_global_state():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    location_US = Location({"country_code": "US"})
    location_IS = Location({"country_code": "IS"})

    assert id(location_US._serializables["country_name"]) == id(
        location_IS._serializables["country_name"])



import pytest

try:
    from bson.objectid import ObjectId
except ImportError:
    ObjectId = None
else:
    from schematics.contrib.mongo import ObjectIdType
    from schematics.exceptions import ConversionError, ValidationError

    FAKE_OID = ObjectId()

pytestmark = pytest.mark.skipif(ObjectId is None,
                                reason='requires pymongo')


def test_to_native():
    oid = ObjectIdType()

    assert oid.to_native(FAKE_OID) == FAKE_OID
    assert oid.to_native(str(FAKE_OID)) == FAKE_OID

    with pytest.raises(ConversionError):
        oid.to_native('foo')


def test_to_primitive():
    oid = ObjectIdType()

    assert oid.to_primitive(FAKE_OID) == str(FAKE_OID)
    assert oid.to_primitive(str(FAKE_OID)) == str(FAKE_OID)


def test_validate_id():
    oid = ObjectIdType()

    assert oid.validate_id(FAKE_OID) is True
    assert oid.validate_id(str(FAKE_OID)) is True

    with pytest.raises(ValidationError):
        oid.validate_id('foo')



# encoding=utf-8

import collections
import itertools

from six import iteritems

from .exceptions import ConversionError, ModelConversionError, ValidationError
from .datastructures import OrderedDict

try:
    basestring #PY2
except NameError:
    basestring = str #PY3

def _list_or_string(lors):
    if lors is None:
        return []
    if isinstance(lors, basestring):
        return [lors]
    return list(lors)

try:
    unicode #PY2
except:
    import codecs
    unicode = str #PY3

###
# Transform Loops
###

def import_loop(cls, instance_or_dict, field_converter, context=None,
                partial=False, strict=False, mapping=None):
    """
    The import loop is designed to take untrusted data and convert it into the
    native types, as described in ``cls``.  It does this by calling
    ``field_converter`` on every field.

    Errors are aggregated and returned by throwing a ``ModelConversionError``.

    :param cls:
        The class for the model.
    :param instance_or_dict:
        A dict of data to be converted into types according to ``cls``.
    :param field_convert:
        This function is applied to every field found in ``instance_or_dict``.
    :param context:
        A ``dict``-like structure that may contain already validated data.
    :param partial:
        Allow partial data to validate; useful for PATCH requests.
        Essentially drops the ``required=True`` arguments from field
        definitions. Default: False
    :param strict:
        Complain about unrecognized keys. Default: False
    """
    is_dict = isinstance(instance_or_dict, dict)
    is_cls = isinstance(instance_or_dict, cls)
    if not is_cls and not is_dict:
        error_msg = 'Model conversion requires a model or dict'
        raise ModelConversionError(error_msg)
    if mapping is None:
        mapping = {}
    data = dict(context) if context is not None else {}
    errors = {}

    # Determine all acceptable field input names
    all_fields = set(cls._fields) ^ set(cls._serializables)
    for field_name, field, in iteritems(cls._fields):
        if hasattr(field, 'serialized_name'):
            all_fields.add(field.serialized_name)
        if hasattr(field, 'deserialize_from'):
            all_fields.update(set(_list_or_string(field.deserialize_from)))
        if field_name in mapping:
            all_fields.update(set(_list_or_string(mapping[field_name])))

    # Check for rogues if strict is set
    rogue_fields = set(instance_or_dict) - set(all_fields)
    if strict and len(rogue_fields) > 0:
        for field in rogue_fields:
            errors[field] = 'Rogue field'

    for field_name, field in iteritems(cls._fields):
        serialized_field_name = field.serialized_name or field_name

        trial_keys = _list_or_string(field.deserialize_from)
        trial_keys.extend(mapping.get(field_name, []))
        trial_keys.extend([serialized_field_name, field_name])

        raw_value = None
        for key in trial_keys:
            if key and key in instance_or_dict:
                raw_value = instance_or_dict[key]
        if raw_value is None:
            if field_name in data:
                continue
            raw_value = field.default

        try:
            if raw_value is None:
                if field.required and not partial:
                    errors[serialized_field_name] = [field.messages['required']]
            else:
                try:
                    mapping_by_model = mapping.get('model_mapping', {})
                    model_mapping = mapping_by_model.get(field_name, {})
                    raw_value = field_converter(field, raw_value, mapping=model_mapping)
                except Exception:
                    raw_value = field_converter(field, raw_value)

            data[field_name] = raw_value

        except ConversionError as exc:
            errors[serialized_field_name] = exc.messages
        except ValidationError as exc:
            errors[serialized_field_name] = exc.messages

    if errors:
        raise ModelConversionError(errors, data)

    return data


def export_loop(cls, instance_or_dict, field_converter,
                role=None, raise_error_on_role=False, print_none=False):
    """
    The export_loop function is intended to be a general loop definition that
    can be used for any form of data shaping, such as application of roles or
    how a field is transformed.

    :param cls:
        The model definition.
    :param instance_or_dict:
        The structure where fields from cls are mapped to values. The only
        expectionation for this structure is that it implements a ``dict``
        interface.
    :param field_converter:
        This function is applied to every field found in ``instance_or_dict``.
    :param role:
        The role used to determine if fields should be left out of the
        transformation.
    :param raise_error_on_role:
        This parameter enforces strict behavior which requires substructures
        to have the same role definition as their parent structures.
    :param print_none:
        This function overrides ``serialize_when_none`` values found either on
        ``cls`` or an instance.
    """
    data = {}

    # Translate `role` into `gottago` function
    gottago = wholelist()
    if hasattr(cls, '_options') and role in cls._options.roles:
        gottago = cls._options.roles[role]
    elif role and raise_error_on_role:
        error_msg = u'%s Model has no role "%s"'
        raise ValueError(error_msg % (cls.__name__, role))
    else:
        gottago = cls._options.roles.get("default", gottago)

    fields_order = (getattr(cls._options, 'fields_order', None)
                    if hasattr(cls, '_options') else None)

    for field_name, field, value in atoms(cls, instance_or_dict):
        serialized_name = field.serialized_name or field_name

        # Skipping this field was requested
        if gottago(field_name, value):
            continue

        # Value found, apply transformation and store it
        elif value is not None:
            if hasattr(field, 'export_loop'):
                shaped = field.export_loop(value, field_converter,
                                           role=role,
                                           print_none=print_none)
            else:
                shaped = field_converter(field, value)

            # Print if we want none or found a value
            if shaped is None and allow_none(cls, field):
                data[serialized_name] = shaped
            elif shaped is not None:
                data[serialized_name] = shaped
            elif print_none:
                data[serialized_name] = shaped

        # Store None if reqeusted
        elif value is None and allow_none(cls, field):
            data[serialized_name] = value
        elif print_none:
            data[serialized_name] = value

    # Return data if the list contains anything
    if len(data) > 0:
        if fields_order:
            return sort_dict(data, fields_order)
        return data
    elif print_none:
        return data


def sort_dict(dct, based_on):
    """
    Sorts provided dictionary based on order of keys provided in ``based_on``
    list.

    Order is not guarantied in case if ``dct`` has keys that are not present
    in ``based_on``

    :param dct:
        Dictionary to be sorted.
    :param based_on:
        List of keys in order that resulting dictionary should have.
    :return:
        OrderedDict with keys in the same order as provided ``based_on``.
    """
    return OrderedDict(
        sorted(
            dct.items(),
            key=lambda el: based_on.index(el[0] if el[0] in based_on else -1))
    )


def atoms(cls, instance_or_dict):
    """
    Iterator for the atomic components of a model definition and relevant data
    that creates a threeple of the field's name, the instance of it's type, and
    it's value.

    :param cls:
        The model definition.
    :param instance_or_dict:
        The structure where fields from cls are mapped to values. The only
        expectionation for this structure is that it implements a ``dict``
        interface.
    """
    all_fields = itertools.chain(iteritems(cls._fields),
                                 iteritems(cls._serializables))

    return ((field_name, field, instance_or_dict[field_name])
            for field_name, field in all_fields)


def allow_none(cls, field):
    """
    This function inspects a model and a field for a setting either at the
    model or field level for the ``serialize_when_none`` setting.

    The setting defaults to the value of the class.  A field can override the
    class setting with it's own ``serialize_when_none`` setting.

    :param cls:
        The model definition.
    :param field:
        The field in question.
    """
    allowed = cls._options.serialize_when_none
    if field.serialize_when_none is not None:
        allowed = field.serialize_when_none
    return allowed


###
# Field Filtering
###

class Role(collections.Set):

    """
    A ``Role`` object can be used to filter specific fields against a sequence.

    The ``Role`` is two things: a set of names and a function.  The function
    describes how filter taking a field name as input and then returning either
    ``True`` or ``False``, indicating that field should or should not be
    skipped.

    A ``Role`` can be operated on as a ``Set`` object representing the fields
    is has an opinion on.  When Roles are combined with other roles, the
    filtering behavior of the first role is used.
    """

    def __init__(self, function, fields):
        self.function = function
        self.fields = set(fields)

    def _from_iterable(self, iterable):
        return Role(self.function, iterable)

    def __contains__(self, value):
        return value in self.fields

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __eq__(self, other):
        print(dir(self.function))
        return (self.function.__name__ == other.function.__name__ and
                self.fields == other.fields)

    def __str__(self):
        return '%s(%s)' % (self.function.__name__,
                           ', '.join("'%s'" % f for f in self.fields))

    def __repr__(self):
        return '<Role %s>' % str(self)

    # edit role fields
    def __add__(self, other):
        fields = self.fields.union(other)
        return self._from_iterable(fields)

    def __sub__(self, other):
        fields = self.fields.difference(other)
        return self._from_iterable(fields)

    # apply role to field
    def __call__(self, name, value):
        return self.function(name, value, self.fields)

    # static filter functions
    @staticmethod
    def wholelist(name, value, seq):
        """
        Accepts a field name, value, and a field list.  This functions
        implements acceptance of all fields by never requesting a field be
        skipped, thus returns False for all input.

        :param name:
            The field name to inspect.
        :param value:
            The field's value.
        :param seq:
            The list of fields associated with the ``Role``.
        """
        return False

    @staticmethod
    def whitelist(name, value, seq):
        """
        Implements the behavior of a whitelist by requesting a field be skipped
        whenever it's name is not in the list of fields.

        :param name:
            The field name to inspect.
        :param value:
            The field's value.
        :param seq:
            The list of fields associated with the ``Role``.
        """

        if seq is not None and len(seq) > 0:
            return name not in seq
        return True

    @staticmethod
    def blacklist(name, value, seq):
        """
        Implements the behavior of a blacklist by requesting a field be skipped
        whenever it's name is found in the list of fields.

        :param k:
            The field name to inspect.
        :param v:
            The field's value.
        :param seq:
            The list of fields associated with the ``Role``.
        """
        if seq is not None and len(seq) > 0:
            return name in seq
        return False


def wholelist(*field_list):
    """
    Returns a function that evicts nothing. Exists mainly to be an explicit
    allowance of all fields instead of a using an empty blacklist.
    """
    return Role(Role.wholelist, field_list)


def whitelist(*field_list):
    """
    Returns a function that operates as a whitelist for the provided list of
    fields.

    A whitelist is a list of fields explicitly named that are allowed.
    """
    return Role(Role.whitelist, field_list)


def blacklist(*field_list):
    """
    Returns a function that operates as a blacklist for the provided list of
    fields.

    A blacklist is a list of fields explicitly named that are not allowed.
    """
    return Role(Role.blacklist, field_list)


###
# Import and export functions
###


def convert(cls, instance_or_dict, context=None, partial=True, strict=False,
            mapping=None):
    def field_converter(field, value, mapping=None):
        try:
            return field.to_native(value, mapping=mapping)
        except Exception:
            return field.to_native(value)
#   field_converter = lambda field, value: field.to_native(value)
    data = import_loop(cls, instance_or_dict, field_converter, context=context,
                       partial=partial, strict=strict, mapping=mapping)
    return data


def to_native(cls, instance_or_dict, role=None, raise_error_on_role=True,
              context=None):
    field_converter = lambda field, value: field.to_native(value,
                                                           context=context)
    data = export_loop(cls, instance_or_dict, field_converter,
                       role=role, raise_error_on_role=raise_error_on_role)
    return data


def to_primitive(cls, instance_or_dict, role=None, raise_error_on_role=True,
                 context=None):
    """
    Implements serialization as a mechanism to convert ``Model`` instances into
    dictionaries keyed by field_names with the converted data as the values.

    The conversion is done by calling ``to_primitive`` on both model and field
    instances.

    :param cls:
        The model definition.
    :param instance_or_dict:
        The structure where fields from cls are mapped to values. The only
        expectionation for this structure is that it implements a ``dict``
        interface.
    :param role:
        The role used to determine if fields should be left out of the
        transformation.
    :param raise_error_on_role:
        This parameter enforces strict behavior which requires substructures
        to have the same role definition as their parent structures.
    """
    field_converter = lambda field, value: field.to_primitive(value,
                                                              context=context)
    data = export_loop(cls, instance_or_dict, field_converter,
                       role=role, raise_error_on_role=raise_error_on_role)
    return data


def serialize(cls, instance_or_dict, role=None, raise_error_on_role=True,
              context=None):
    return to_primitive(cls, instance_or_dict, role, raise_error_on_role,
                        context)


EMPTY_LIST = "[]"
EMPTY_DICT = "{}"


def expand(data, context=None):
    """
    Expands a flattened structure into it's corresponding layers.  Essentially,
    it is the counterpart to ``flatten_to_dict``.

    :param data:
        The data to expand.
    :param context:
        Existing expanded data that this function use for output
    """
    expanded_dict = {}

    if context is None:
        context = expanded_dict

    for key, value in iteritems(data):
        try:
            key, remaining = key.split(".", 1)
        except ValueError:
            if not (value in (EMPTY_DICT, EMPTY_LIST) and key in expanded_dict):
                expanded_dict[key] = value
        else:
            current_context = context.setdefault(key, {})
            if current_context in (EMPTY_DICT, EMPTY_LIST):
                current_context = {}
                context[key] = current_context

            current_context.update(expand({remaining: value}, current_context))
    return expanded_dict


def flatten_to_dict(instance_or_dict, prefix=None, ignore_none=True):
    """
    Flattens an iterable structure into a single layer dictionary.

    For example:

        {
            's': 'jms was hrrr',
            'l': ['jms was here', 'here', 'and here']
        }

        becomes

        {
            's': 'jms was hrrr',
            u'l.1': 'here',
            u'l.0': 'jms was here',
            u'l.2': 'and here'
        }

    :param instance_or_dict:
        The structure where fields from cls are mapped to values. The only
        expectionation for this structure is that it implements a ``dict``
        interface.
    :param ignore_none:
        This ignores any ``serialize_when_none`` settings and forces the empty
        fields to be printed as part of the flattening.
        Default: True
    :param prefix:
        This puts a prefix in front of the field names during flattening.
        Default: None
    """
    if isinstance(instance_or_dict, dict):
        iterator = iteritems(instance_or_dict)
    # if hasattr(instance_or_dict, "iteritems"):
    #     iterator = instance_or_dict.iteritems()
    else:
        iterator = enumerate(instance_or_dict)

    flat_dict = {}
    for key, value in iterator:
        if prefix:
            key = ".".join(map(unicode, (prefix, key)))

        if value == []:
            value = EMPTY_LIST
        elif value == {}:
            value = EMPTY_DICT

        if isinstance(value, (dict, list)):
            flat_dict.update(flatten_to_dict(value, prefix=key))
        elif value is not None:
            flat_dict[key] = value
        elif not ignore_none:
            flat_dict[key] = None

    return flat_dict


def flatten(cls, instance_or_dict, role=None, raise_error_on_role=True,
            ignore_none=True, prefix=None, context=None):
    """
    Produces a flat dictionary representation of the model.  Flat, in this
    context, means there is only one level to the dictionary.  Multiple layers
    are represented by the structure of the key.

    Example:

        >>> class Foo(Model):
        ...    s = StringType()
        ...    l = ListType(StringType)

        >>> f = Foo()
        >>> f.s = 'string'
        >>> f.l = ['jms', 'was here', 'and here']

        >>> flatten(Foo, f)
        {'s': 'string', u'l.1': 'jms', u'l.0': 'was here', u'l.2': 'and here'}

    :param cls:
        The model definition.
    :param instance_or_dict:
        The structure where fields from cls are mapped to values. The only
        expectionation for this structure is that it implements a ``dict``
        interface.
    :param role:
        The role used to determine if fields should be left out of the
        transformation.
    :param raise_error_on_role:
        This parameter enforces strict behavior which requires substructures
        to have the same role definition as their parent structures.
    :param ignore_none:
        This ignores any ``serialize_when_none`` settings and forces the empty
        fields to be printed as part of the flattening.
        Default: True
    :param prefix:
        This puts a prefix in front of the field names during flattening.
        Default: None
    """
    field_converter = lambda field, value: field.to_primitive(value,
                                                              context=context)

    data = export_loop(cls, instance_or_dict, field_converter,
                       role=role, print_none=True)

    flattened = flatten_to_dict(data, prefix=prefix, ignore_none=ignore_none)

    return flattened



from copy import deepcopy
from six.moves import zip
from six import iteritems
from six import PY3

_missing = object()


class OrderedDict(dict):

    """Simple ordered dict implementation.

    It's a dict subclass and provides some list functions.  The implementation
    of this class is inspired by the implementation of Babel but incorporates
    some ideas from the `ordereddict`_ and Django's ordered dict.

    The constructor and `update()` both accept iterables of tuples as well as
    mappings:

    >>> d = OrderedDict([('a', 'b'), ('c', 'd')])
    >>> d.update({'foo': 'bar'})
    >>> d
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])

    Keep in mind that when updating from dict-literals the order is not
    preserved as these dicts are unsorted!

    You can copy an OrderedDict like a dict by using the constructor,
    `copy.copy` or the `copy` method and make deep copies with `copy.deepcopy`:

    >>> from copy import copy, deepcopy
    >>> copy(d)
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    >>> d.copy()
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    >>> OrderedDict(d)
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar')])
    >>> d['spam'] = []
    >>> d2 = deepcopy(d)
    >>> d2['spam'].append('eggs')
    >>> d
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar'), ('spam', [])])
    >>> d2
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar'), ('spam', ['eggs'])])

    All iteration methods as well as `keys`, `values` and `items` return
    the values ordered by the the time the key-value pair is inserted:

    >>> d.keys()
    ['a', 'c', 'foo', 'spam']
    >>> d.values()
    ['b', 'd', 'bar', []]
    >>> d.items()
    [('a', 'b'), ('c', 'd'), ('foo', 'bar'), ('spam', [])]
    >>> list(d.iterkeys())
    ['a', 'c', 'foo', 'spam']
    >>> list(d.itervalues())
    ['b', 'd', 'bar', []]
    >>> list(d.iteritems())
    [('a', 'b'), ('c', 'd'), ('foo', 'bar'), ('spam', [])]

    You can sort the OrderedDict like a list:

    >>> d.sort(key=lambda x: x[0].lower())
    >>> d
    OrderedDict([('a', 'b'), ('c', 'd'), ('foo', 'bar'), ('spam', [])])

    For performance reasons the ordering is not taken into account when
    comparing two ordered dicts.

    .. _ordereddict: http://www.xs4all.nl/~anthon/Python/ordereddict/
    """

    def __init__(self, *args, **kwargs):
        super(OrderedDict, self).__init__()
        self._keys = []
        self.update(*args, **kwargs)

    def __delitem__(self, key):
        super(OrderedDict, self).__delitem__(key)
        self._keys.remove(key)

    def __setitem__(self, key, item):
        if key not in self:
            if hasattr(self, '_keys'):
                self._keys.append(key)
            else:
                self._keys = [key]
        super(OrderedDict, self).__setitem__(key, item)

    def __deepcopy__(self, memo):
        memo[id(self)] = new_od = self.__class__()
        new_od.__init__(deepcopy(self.items(), memo))
        return new_od

    def __reversed__(self):
        return reversed(self._keys)

    @classmethod
    def fromkeys(cls, iterable, default=None):
        return cls((key, default) for key in iterable)

    def clear(self):
        del self._keys[:]
        super(OrderedDict, self).clear()

    def copy(self):
        return self.__class__(self)

    def items(self):
        return list(zip(self._keys, self.values()))

    def iteritems(self):
        return list(zip(self._keys, self.itervalues()))

    def keys(self):
        return self._keys[:]

    def iterkeys(self):
        return iter(self._keys)

    def pop(self, key, default=_missing):
        if key not in self:
            if default is _missing:
                raise KeyError(key)
            return default
        self._keys.remove(key)
        return super(OrderedDict, self).pop(key, default)

    def popitem(self):
        if not self._keys:
            raise KeyError('popitem(): dictionary is empty')
        return self._keys[0], self.pop(self._keys[0])

    def setdefault(self, key, default=None):
        if key not in self:
            self._keys.append(key)
        return super(OrderedDict, self).setdefault(key, default)

    def update(self, *args, **kwargs):
        sources = []
        if len(args) == 1:
            if isinstance(args[0], dict):
                sources.append(iteritems(args[0]))
            # if hasattr(args[0], 'iteritems'):
            #     sources.append(args[0].iteritems())
            else:
                sources.append(iter(args[0]))
        elif args:
            raise TypeError('expected at most one positional argument')
        if kwargs:
            sources.append(iteritems(kwargs))
        for iterable in sources:
            for key, val in iterable:
                self[key] = val

    def values(self):
        return [self.get(key) for key in self._keys]

    def itervalues(self):
        return (self.get(key) for key in self._keys)

    def sort(self, cmp=None, key=None, reverse=False):
        if key is not None:
            self._keys.sort(key=lambda k: key((k, self[k])))
        elif cmp is not None:
            self._keys.sort(lambda a, b: cmp((a, self[a]), (b, self[b])))
        else:
            self._keys.sort()
        if reverse:
            self._keys.reverse()

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.items())

    __copy__ = copy
    __iter__ = iterkeys


class DataObject(object):
    """
    An object for holding data as attributes.

    ``DataObject`` can be instantiated like ``dict``::

        >>> d = DataObject({'one': 1, 'two': 2}, three=3)
        >>> d.__dict__
        {'one': 1, 'two': 2, 'three': 3}

    Attributes are accessible via the regular dot notation (``d.x``) as well as
    the subscription syntax (``d['x']``)::

        >>> d.one == d['one'] == 1
        True

    Dictionaries stored as attributes will be automatically converted into
    ``DataObject`` instances. Nested dictionaries will be converted recursively::

        >>> e = DataObject({'f': {'g': {'x': 1, 'y': 2}}})
        >>> e.f.g.x
        1

    To convert a ``DataObject`` structure into a dictionary, use ``d._to_dict()``.

    ``DataObject`` implements the following collection-like operations:

        * iteration through attributes as name-value pairs
        * ``'x' in d`` for membership tests
        * ``len(d)`` to get the number of attributes

    Additionally, the following methods are equivalent to their ``dict` counterparts:
    ``_clear``, ``_get``, ``_items``, ``_pop``, ``_setdefault``, ``_update``.

    An advantage of ``DataObject`` over ``dict` subclasses is that every method name
    in ``DataObject`` begins with an underscore, so attributes like ``"update"`` or
    ``"values"`` are valid.
    """

    def __init__(self, *args, **kwargs):
        source = args[0] if args else {}
        self._update(source, **kwargs)

    def __setattr__(self, name, value):
        if isinstance(value, dict):
            value = self.__class__(value)
        self.__dict__[name] = value

    __setitem__ = __setattr__

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % repr(self.__dict__)

    def _copy(self):
        return self.__class__(self)

    __copy__ = _copy

    def __eq__(self, other):
        return isinstance(other, DataObject) and self.__dict__ == other.__dict__

    def __iter__(self):
        return iter(self.__dict__.items())

    def _update(self, source=(), **kwargs):
        if isinstance(source, dict):
            source = source.items()
        for k, v in source:
            self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def _setdefault(self, name, value=None):
        if name not in self:
            self[name] = value
        return self[name]

    def _setdefaults(self, source):
        if isinstance(source, dict):
            source = source.items()
        for k, v in source:
            self._setdefault(k, v)

    def _to_dict(self):
        d = dict(self.__dict__)
        for k, v in d.items():
            if isinstance(v, DataObject):
                d[k] = v._to_dict()
        return d

    def __getitem__(self, key): return self.__dict__[key]
    def __delitem__(self, key): del self.__dict__[key]
    def __len__(self): return len(self.__dict__)
    def __contains__(self, key): return key in self.__dict__

    def _clear(self): return self.__dict__.clear()
    def _get(self, *args): return self.__dict__.get(*args)
    def _items(self): return self.__dict__.items()
    def _pop(self, *args): return self.__dict__.pop(*args)


class ConfigObject(DataObject):
    """
    A variant of ``DataObject`` that returns ``None`` by default when a nonexistent
    attribute is requested. That is, ``d.x`` is equivalent to ``d._get('x')``.
    """

    def __getattr__(self, name):
        return None

    __getitem__ = __getattr__




import datetime
import decimal
import uuid

import pytest

from schematics.types import (
    BaseType, StringType, DateTimeType, DateType, IntType, EmailType, LongType,
    URLType, MultilingualStringType, UUIDType, IPv4Type, MD5Type, BooleanType,
    GeoPointType, FloatType, DecimalType, force_unicode
)
from schematics.types.base import get_range_endpoints
from schematics.exceptions import (
    ValidationError, ConversionError, MockCreationError
)


def test_string_choices():
    with pytest.raises(TypeError):
        BaseType(choices='foo')

def test_int():
    with pytest.raises(ConversionError):
        IntType().validate('foo')

    assert IntType.validate(5001) == None


def test_date():
    today = datetime.date(2013, 3, 1)

    date_type = DateType()
    assert date_type("2013-03-01") == today

    assert date_type.to_primitive(today) == "2013-03-01"

    assert date_type.to_native(today) is today

    with pytest.raises(ConversionError):
        date_type.to_native('foo')


def test_datetime():
    dt = datetime.datetime.now()
    assert DateTimeType()(dt.isoformat()) == dt

    obj = DateTimeType(formats='foo')
    assert obj.formats == ['foo']


def test_datetime_format():
    dt_input = '2013.03.07 15:31'
    dt = datetime.datetime(2013, 3, 7, 15, 31)
    assert DateTimeType(formats=['%Y.%m.%d %H:%M'])(dt_input) == dt


def test_datetime_to_native():
    with pytest.raises(ConversionError):
        DateTimeType().to_native('foo')


def test_datetime_primitive():
    output = '2013.03.07 15:31'
    dt = datetime.datetime(2013, 3, 7, 15, 31)
    dt_type = DateTimeType(serialized_format='%Y.%m.%d %H:%M')
    assert dt_type(dt.isoformat()) == dt
    assert dt_type.to_primitive(dt) == output

    obj = DateTimeType(serialized_format=str)
    assert obj.to_primitive(123) == '123'


def test_datetime_accepts_datetime():
    output = '2013.03.07 15:31'
    dt = datetime.datetime(2013, 3, 7, 15, 31)
    dt_type = DateTimeType(serialized_format='%Y.%m.%d %H:%M')
    assert dt_type(dt) == dt
    assert dt_type.to_primitive(dt) == output


def test_int():
    with pytest.raises(ConversionError):
        IntType()('a')
    assert IntType()(1) == 1


def test_custom_validation_functions():
    class UppercaseType(BaseType):

        def validate_uppercase(self, value):
            if value.upper() != value:
                raise ValidationError("Value must be uppercase!")

    field = UppercaseType()

    with pytest.raises(ValidationError):
        field.validate("lowercase")


def test_custom_validation_function_and_inheritance():
    class UppercaseType(StringType):

        def validate_uppercase(self, value):
            if value.upper() != value:
                raise ValidationError("Value must be uppercase!")

    class MUppercaseType(UppercaseType):

        def __init__(self, number_of_m_chars, **kwargs):
            self.number_of_m_chars = number_of_m_chars

            super(MUppercaseType, self).__init__(**kwargs)

        def validate_contains_m_chars(self, value):
            if value.count("M") != self.number_of_m_chars:
                raise ValidationError(
                    "Value must contain {0} 'm' characters".format(self.number_of_m_chars))

    field = MUppercaseType(number_of_m_chars=3)

    field.validate("MMM")

    with pytest.raises(ValidationError):
        field.validate("mmm")

    with pytest.raises(ValidationError):
        field.validate("MM")


def test_email_type_with_invalid_email():
    with pytest.raises(ValidationError):
        EmailType().validate(u'sdfg\U0001f636\U0001f46e')


def test_url_type_with_invalid_url():
    with pytest.raises(ValidationError):
        URLType().validate(u'http:example.com')


def test_url_type_with_unreachable_url():
    with pytest.raises(ValidationError):
        URLType(verify_exists=True).validate_url('http://127.0.0.1:99999/')


def test_string_type_required():
    field = StringType(required=True)
    with pytest.raises(ValidationError):
        field.validate(None)


def test_string_type_accepts_none():
    field = StringType()
    field.validate(None)


def test_string_required_accepts_empty_string():
    field = StringType(required=True)
    field.validate('')


def test_string_min_length_doesnt_accept_empty_string():
    field = StringType(min_length=1)
    with pytest.raises(ValidationError):
        field.validate('')


def test_string_regex():
    StringType(regex='\d+').validate("1")

    with pytest.raises(ValidationError):
        StringType(regex='\d+').validate("a")


def test_string_to_native():
    with pytest.raises(ConversionError):
        StringType().to_native(3.14)


def test_string_max_length_is_enforced():
    with pytest.raises(ValidationError):
        StringType(max_length=2).validate_length('foo')


def test_multilingualstring_should_only_take_certain_types():
    mls = MultilingualStringType()

    mls(None)
    mls({})

    with pytest.raises(ValueError):
        mls(123)

    with pytest.raises(ValueError):
        mls([])

    with pytest.raises(ValueError):
        mls('foo')


def test_multilingualstring_should_validate_length():
    MultilingualStringType(min_length=3).validate({'en_US': 'foo'})
    MultilingualStringType(max_length=3).validate({'en_US': 'foo'})

    with pytest.raises(ValidationError):
        MultilingualStringType(min_length=4).validate({'en_US': 'foo'})

    with pytest.raises(ValidationError):
        MultilingualStringType(max_length=2).validate({'en_US': 'foo'})


def test_multilingualstring_should_validate_regex():
    MultilingualStringType(regex='^[a-z]*$').validate({'en_US': 'foo'})

    with pytest.raises(ValidationError):
        MultilingualStringType(regex='^[a-z]*$').validate({'en_US': '123'})

    with pytest.raises(ValidationError):
        MultilingualStringType(locale_regex='^\d*$').validate({'en_US': 'foo'})

    assert MultilingualStringType(locale_regex=None).validate_regex({'en_US': 'foo'}) is None


def test_multilingualstring_should_handle_none():
    mls = MultilingualStringType(default_locale='en_US')

    assert mls.to_primitive(None) is None


def test_multilingualstring_should_handle_castable_values():
    mls = MultilingualStringType(default_locale='en_US')

    assert mls.to_primitive({'en_US': 123}) == u'123'


def test_multilingualstring_should_enforce_noncastable_values():
    mls = MultilingualStringType(default_locale='en_US')

    with pytest.raises(ConversionError):
        mls.to_primitive({'en_US': 123.0}) == u'123'


def test_multilingual_string_should_emit_string_with_default_locale():
    mls = MultilingualStringType(default_locale='en_US')

    assert mls.to_primitive({'en_US': 'snake', 'fr_FR': 'serpent'}) == 'snake'


def test_multilingual_string_should_emit_string_with_explicit_locale():
    mls = MultilingualStringType(default_locale='en_US')

    assert mls.to_primitive(
        {'en_US': 'snake', 'fr_FR': 'serpent'},
        context={'locale': 'fr_FR'}) == 'serpent'


def test_multilingual_string_should_require_a_locale():
    mls = MultilingualStringType()

    with pytest.raises(ConversionError):
        mls.to_primitive({'foo': 'bar'})


def test_multilingual_string_without_matching_locale_should_explode():
    mls = MultilingualStringType()

    with pytest.raises(ConversionError):
        mls.to_primitive({'fr_FR': 'serpent'})

    with pytest.raises(ConversionError):
        mls.to_primitive({'en_US': 'snake'}, context={'locale': 'fr_FR'})


def test_multilingual_string_should_accept_lists_of_locales():
    strings = {
        'en_US': 'snake',
        'fr_FR': 'serpent',
        'es_MX': 'serpiente',
    }

    mls = MultilingualStringType(default_locale=['foo', 'fr_FR', 'es_MX'])

    assert mls.to_primitive(strings) == 'serpent'
    assert mls.to_primitive(strings, context={'locale': ['es_MX', 'bar']}) == 'serpiente'

    mls = MultilingualStringType()

    assert mls.to_primitive(strings, context={'locale': ['foo', 'es_MX', 'fr_FR']}) == 'serpiente'


def test_boolean_to_native():
    bool_field = BooleanType()

    for true_value in ['True', '1', 1, True]:
        assert bool_field.to_native(true_value)

    for false_value in ['False', '0', 0, False]:
        assert not bool_field.to_native(false_value)

    for bad_value in ['TrUe', 'foo', 2, None, 1.0]:
        with pytest.raises(ConversionError):
            bool_field.to_native(bad_value)



from datetime import datetime

import pytest

try:
    from dateutil.tz import gettz, tzutc
except ImportError:
    gettz = tzutc = None
else:
    from schematics.types.temporal import TimeStampType

    EPOCH = datetime(1970, 1, 1, 0, 0, tzinfo=tzutc())

pytestmark = pytest.mark.skipif(tzutc is None,
                                reason='requires python-dateutil')


def test_timezone_to_date():
    date = TimeStampType.timestamp_to_date(0)
    assert date == EPOCH

    date = TimeStampType.timestamp_to_date(1399588840)
    assert date == datetime(2014, 5, 8, 22, 40, 40, tzinfo=tzutc())


def test_date_to_timestamp():
    assert TimeStampType.date_to_timestamp(EPOCH) == 0

    ts = TimeStampType.date_to_timestamp(datetime(2014, 5, 8, 22, 40, 40, tzinfo=tzutc()))
    assert ts == 1399588840.0

    ts = TimeStampType.date_to_timestamp(
        datetime(2014, 5, 8, 22, 40, 40, tzinfo=gettz('PST8PDT')))
    assert ts == 1399614040.0


def test_date_to_primitive():
    tsobj = TimeStampType()

    assert tsobj.to_primitive(EPOCH) == 0

    ts = tsobj.to_primitive(datetime(2014, 5, 8, 22, 40, 40, tzinfo=tzutc()))
    assert ts == 1399588840.0

    ts = tsobj.to_primitive(datetime(2014, 5, 8, 22, 40, 40, tzinfo=gettz('PST8PDT')))
    assert ts == 1399614040.0



import pytest

from schematics.models import Model
from schematics.types import StringType
from schematics.types.compound import PolyModelType


class A(Model): # fallback model (doesn't define a claim method)
    stringA = StringType()

class Aaa(A):
    pass

class B(A):
    stringB = StringType()
    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get('stringB') == 'bbb'

class C(B):
    stringC = StringType()
    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get('stringC') == 'ccc'

class X(Model):
    pass

def claim_func(field, data):
    if 'stringB' in data and field.name == 'cfn':
        return B
    if 'stringC' in data and field.name == 'cfn':
        return C
    else:
        return None


class Foo(Model):
    base   = PolyModelType(A)       # accepts any subclass for import and export
    strict = PolyModelType([A, B])  # accepts [A, B] for import and export
    nfb    = PolyModelType([B, C])  # no fallback since A not present
    cfn    = PolyModelType([B, C], claim_function=claim_func, strict=False)


def test_subclass_registry():

    assert A._subclasses == [Aaa, B, C]
    assert B._subclasses == [C]
    assert C._subclasses == []

def test_inheritance_based_polymorphic(): # base

    foo = Foo({'base': {'stringB': 'bbb'}})
    assert type(foo.base) is B

    foo = Foo({'base': {'stringC': 'ccc'}})
    assert type(foo.base) is C

    foo = Foo({'base': {}}) # unrecognizable input => fall back to A
    assert type(foo.base) is A

    foo = Foo({'base': Aaa()})
    assert type(foo.base) is Aaa
    foo.validate()
    foo.to_primitive()

def test_enumerated_polymorphic(): # strict

    foo = Foo({'strict': {'stringB': 'bbb'}})
    assert type(foo.strict) is B

    foo = Foo({'strict': {}}) # unrecognizable input => fall back to A
    assert type(foo.strict) is A

    foo = Foo({'strict': B()})
    assert type(foo.strict) is B
    foo.validate()
    foo.to_primitive()

    Foo.strict.allow_subclasses = True
    foo = Foo({'strict': Aaa()})
    Foo.strict.allow_subclasses = False

def test_external_claim_function(): # cfn

    foo = Foo({'cfn': {'stringB': 'bbb', 'stringC': 'ccc'}})
    assert type(foo.cfn) is B

def test_multiple_matches():

    with pytest.raises(Exception):
        foo = Foo({'base': {'stringB': 'bbb', 'stringC': 'ccc'}})

def test_no_fallback(): # nfb

    with pytest.raises(Exception):
        foo = Foo({'nfb': {}})

def test_refuse_unrelated_import():

    with pytest.raises(Exception):
        foo = Foo({'base': D()})

    with pytest.raises(Exception):
        foo = Foo({'strict': Aaa()})

def test_refuse_unrelated_export():

    with pytest.raises(Exception):
        foo = Foo()
        foo.base = X()
        foo.to_primitive()

    with pytest.raises(Exception):
        foo = Foo()
        foo.strict = Aaa()
        foo.to_primitive()




import pytest
import datetime

from schematics.models import Model
from schematics.exceptions import (
    BaseError, ValidationError, ConversionError,
    ModelValidationError, ModelConversionError,
)
from schematics.types import StringType, DateTimeType, BooleanType
from schematics.types.compound import ModelType, ListType, DictType


future_error_msg = u'Future dates are not valid'


def test_choices_validates():
    class Document(Model):
        language = StringType(choices=['en', 'de'])

    doc = Document({'language': 'de'})
    doc.validate()


def test_validation_fails():
    class Document(Model):
        language = StringType(choices=['en', 'de'])

    doc = Document({'language': 'fr'})
    with pytest.raises(ValidationError):
        doc.validate()


def test_choices_validates_with_embedded():
    class Document(Model):
        language = StringType(choices=['en', 'de'])
        other = ListType(StringType())

    doc = Document({
        'language': 'de',
        'other': ['somevalue', 'other']
    })

    doc.validate()


def test_validation_failes_with_embedded():
    class Document(Model):
        language = StringType(choices=['en', 'de'])
        other = ListType(StringType())

    doc = Document({
        'language': 'fr',
        'other': ['somevalue', 'other']
    })

    with pytest.raises(ValidationError):
        doc.validate()


def test_validation_none_fails():
    class Player(Model):
        first_name = StringType(required=True)

    with pytest.raises(ValidationError) as exception:
        Player({"first_name": None}).validate()

        assert len(exception.messages) == 1  # Only one failure
        assert u'This field is required' in exception.messages["first_name"][0]


def test_custom_validators():
    now = datetime.datetime(2012, 1, 1, 0, 0)

    def is_not_future(dt, *args):
        if dt > now:
            raise ValidationError(future_error_msg)

    class TestDoc(Model):
        publish = DateTimeType(validators=[is_not_future])
        author = StringType(required=True)
        title = StringType(required=False)

    with pytest.raises(ValidationError) as exception:
        TestDoc({
            "publish": datetime.datetime(2012, 2, 1, 0, 0),
            "author": "Hemingway",
            "title": "Old Man"
        }).validate()

        assert future_error_msg in exception.messages['publish']


def test_messages_subclassing():
    class MyStringType(StringType):
        MESSAGES = {'required': u'Never forget'}

    class TestDoc(Model):
        title = MyStringType(required=True)

    with pytest.raises(ValidationError) as exception:
        TestDoc({'title': None}).validate()

        assert u'Never forget' in exception.messages['title']


def test_messages_instance_level():
    class TestDoc(Model):
        title = StringType(required=True, messages={'required': u'Never forget'})

    with pytest.raises(ValidationError) as exception:
        TestDoc({'title': None}).validate()
        assert u'Never forget' in exception.messages['title']


def test_model_validators():
    now = datetime.datetime(2012, 1, 1, 0, 0)
    future = now + datetime.timedelta(1)

    assert future > now

    class TestDoc(Model):
        can_future = BooleanType()
        publish = DateTimeType()

        def validate_publish(self, data, dt):
            if dt > datetime.datetime(2012, 1, 1, 0, 0) and not data['can_future']:
                raise ValidationError(future_error_msg)

    with pytest.raises(ValidationError):
        TestDoc({'publish': future}).validate()


def test_multi_key_validation():
    now = datetime.datetime(2012, 1, 1, 0, 0)
    future = now + datetime.timedelta(1)

    input = str(future.isoformat())

    class GoodModel(Model):
        should_raise = BooleanType(default=True)
        publish = DateTimeType()

        def validate_publish(self, data, dt):
            if data['should_raise'] is True:
                raise ValidationError(u'')
            return dt

    with pytest.raises(ValidationError):
        GoodModel({'publish': input}).validate()

    GoodModel({'publish': input, 'should_raise': False}).validate()

    with pytest.raises(ValidationError):
        GoodModel({'publish': input, 'should_raise': True}).validate()


def test_multi_key_validation_part_two():
    class Signup(Model):
        name = StringType()
        call_me = BooleanType(default=False)

        def validate_call_me(self, data, value):
            if data['name'] == u'Brad' and value is True:
                raise ValidationError(u'I\'m sorry I never call people who\'s name is Brad')
            return value

    Signup({'name': u'Brad'}).validate()
    Signup({'name': u'Brad', 'call_me': False}).validate()

    with pytest.raises(ValidationError):
        Signup({'name': u'Brad', 'call_me': True}).validate()


def test_multi_key_validation_fields_order():
    class Signup(Model):
        name = StringType()
        call_me = BooleanType(default=False)

        def validate_name(self, data, value):
            if data['name'] == u'Brad':
                value = u'Joe'
                data['name'] = value
                return value
            return value

        def validate_call_me(self, data, value):
            if data['name'] == u'Joe':
                raise ValidationError(u"Don't try to decept me! You're Joe!")
            return value

    Signup({'name': u'Tom'}).validate()

    with pytest.raises(ValidationError):
        Signup({'name': u'Brad'}).validate()



def test_basic_error():
    class School(Model):
        name = StringType(required=True)

    school = School()

    with pytest.raises(ValidationError) as exception:
        school.validate()

        errors = exception.messages

        assert "name" in errors
        assert errors["name"] == ["This field is required."]


def test_deep_errors():
    class Person(Model):
        name = StringType(required=True)

    class School(Model):
        name = StringType(required=True)
        headmaster = ModelType(Person, required=True)

    school = School()
    school.name = "Hogwarts"
    school.headmaster = {}

    with pytest.raises(ValidationError) as exception:
        school.validate()

        errors = exception.messages

        assert "headmaster" in errors
        assert "name" in errors["headmaster"]
        assert errors["headmaster"]["name"] == ["This field is required."]


def test_deep_errors_with_lists():
    class Person(Model):
        name = StringType(required=True)

    class Course(Model):
        id = StringType(required=True, validators=[])
        attending = ListType(ModelType(Person))

    class School(Model):
        courses = ListType(ModelType(Course))

    valid_data = {
        'courses': [
            {'id': 'ENG103', 'attending': [
                {'name': u'Danny'},
                {'name': u'Sandy'}]},
            {'id': 'ENG203', 'attending': [
                {'name': u'Danny'},
                {'name': u'Sandy'}
            ]}
        ]
    }

    school = School(valid_data)
    school.validate()

    invalid_data = school.serialize()
    invalid_data['courses'][0]['attending'][0]['name'] = None

    school = School(invalid_data)
    with pytest.raises(ValidationError) as exception:
        school.validate()

        messages = exception.messages

        assert messages == {
            'courses': [
                {
                    'attending': [
                        {
                            'name': [u'This field is required.'],
                        },
                    ],
                }
            ]
        }


def test_deep_errors_with_dicts():
    class Person(Model):
        name = StringType(required=True)

    class Course(Model):
        id = StringType(required=True, validators=[])
        attending = ListType(ModelType(Person))

    class School(Model):
        courses = DictType(ModelType(Course))

    valid_data = {
        'courses': {
            "ENG103": {
                'id': 'ENG103', 'attending': [
                    {'name': u'Danny'},
                    {'name': u'Sandy'},
                ],
            },
            "ENG203": {
                'id': 'ENG203', 'attending': [
                    {'name': u'Danny'},
                    {'name': u'Sandy'},
                ],
            },
        }
    }

    school = School(valid_data)
    school.validate()

    invalid_data = school.serialize()
    invalid_data['courses']["ENG103"]['attending'][0]['name'] = None

    school = School(invalid_data)
    with pytest.raises(ValidationError) as exception:
        school.validate()

        messages = exception.messages

        assert messages == {
            'courses': {
                "ENG103":
                {
                    'attending': [
                        {
                            'name': [u'This field is required.']
                        }
                    ]
                }
            }
        }


def test_clean_validation_messages():
    error = BaseError(["A"])
    assert error.messages == ["A"]


def test_clean_validation_messages_list():
    error = BaseError(["A", "B", "C"])
    assert error.messages, ["A", "B", "C"]


def test_clean_validation_messages_dict():
    error = BaseError({"A": "B"})
    assert error.messages == {"A": "B"}


def test_builtin_conversion_exception():
    with pytest.raises(TypeError):
        raise ConversionError('TypeError')

    with pytest.raises(TypeError):
        raise ModelConversionError('TypeError')


def test_builtin_validation_exception():
    with pytest.raises(ValueError):
        raise ValidationError('ValueError')

    with pytest.raises(ValueError):
        raise ModelValidationError('ValueError')



# -*- coding: utf-8 -*-
from schematics.datastructures import OrderedDict
from schematics.models import Model
from schematics.types import StringType, IntType


def test_serialize_field_orders_one_two_three():
    class Example(Model):
        one = StringType()
        two = IntType()
        three = StringType()

        class Options:
            fields_order = ['one', 'two', 'three']

    p1 = Example({'one': 'a', 'two': 2, 'three': '3'})
    order = [('one', 'a'), ('two', 2), ('three', '3')]
    serialized = p1.to_native()
    assert serialized.items() == order


def test_serialize_field_orders_three_two_one():
    class Example(Model):
        one = StringType()
        two = IntType()
        three = StringType()

        class Options:
            fields_order = ['three', 'two', 'one']

    p1 = Example({'one': 'a', 'two': 2, 'three': '3'})
    serialized = p1.to_native()
    order = [('three', '3'), ('two', 2), ('one', 'a')]
    assert serialized.items() == order


def test_serialize_field_orders_one_three_two():
    class Example(Model):
        one = StringType()
        two = IntType()
        three = StringType()

        class Options:
            fields_order = ['one', 'three', 'two']

    p1 = Example({'one': 'a', 'two': 2, 'three': '3'})
    serialized = p1.to_native()
    order = [('one', 'a'), ('three', '3'), ('two', 2)]
    assert serialized.items() == order



import datetime

try:
    from dateutil.tz import tzutc, tzlocal
except ImportError:  # pragma: no cover
    raise ImportError(
        'Using the datetime fields requires the dateutil library. '
        'You can obtain dateutil from http://labix.org/python-dateutil'
    )

from .base import DateTimeType

EPOCH = datetime.datetime.utcfromtimestamp(0).replace(tzinfo=tzutc())


class TimeStampType(DateTimeType):

    """Variant of a datetime field that saves itself as a unix timestamp (int)
    instead of a ISO-8601 string.
    """

    @classmethod
    def timestamp_to_date(cls, value):
        return datetime.datetime.fromtimestamp(value, tz=tzutc())

    @classmethod
    def date_to_timestamp(cls, value):
        if value.tzinfo is None:
            value = value.replace(tzinfo=tzlocal())
        delta = value - EPOCH
        return (delta.days * 24 * 3600) + delta.seconds + delta.microseconds

    def to_primitive(self, value, context=None):
        return TimeStampType.date_to_timestamp(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand

from schematics import __version__


class Tox(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--recreate']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


tests_require = open(
    os.path.join(os.path.dirname(__file__), 'requirements.txt')).read().split()

setup(
    name='schematics',
    license='BSD',
    version=__version__,
    description='Structured Data for Humans',
    author=u'James Dennis, Jökull Sólberg, Jóhann Þorvaldur Bergþórsson',
    author_email='jdennis@gmail.com, jokull@plainvanillagames.com, johann@plainvanillagames.com',
    url='http://github.com/schematics/schematics',
    packages=['schematics', 'schematics.types', 'schematics.contrib'],
    classifiers=[
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: Other/Proprietary License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    cmdclass={
        'test': Tox,
    },
    install_requires=[
        'six>=1.7.3',
    ],
    tests_require=tests_require,
)



# -*- coding: utf-8 -*-

from __future__ import division

from collections import Iterable
import itertools

from ..exceptions import ValidationError, ConversionError, ModelValidationError, StopValidation
from ..transforms import export_loop, EMPTY_LIST, EMPTY_DICT
from .base import BaseType

from six import iteritems
from six import string_types as basestring
from six import text_type as unicode

class MultiType(BaseType):

    def validate(self, value):
        """Report dictionary of errors with lists of errors as values of each
        key. Used by ModelType and ListType.

        """
        errors = {}

        for validator in self.validators:
            try:
                validator(value)
            except ModelValidationError as exc:
                errors.update(exc.messages)
            except StopValidation as exc:
                errors.update(exc.messages)
                break

        if errors:
            raise ValidationError(errors)

        return value

    def export_loop(self, shape_instance, field_converter,
                    role=None, print_none=False):
        raise NotImplementedError

    def init_compound_field(self, field, compound_field, **kwargs):
        """
        Some of non-BaseType fields requires `field` arg.
        Not avoid name conflict, provide it as `compound_field`.
        Example:

            comments = ListType(DictType, compound_field=StringType)
        """
        if compound_field:
            field = field(field=compound_field, **kwargs)
        else:
            field = field(**kwargs)
        return field


class ModelType(MultiType):

    def __init__(self, model_class, **kwargs):
        self.model_class = model_class
        self.fields = self.model_class.fields

        validators = kwargs.pop("validators", [])
        self.strict = kwargs.pop("strict", True)

        def validate_model(model_instance):
            model_instance.validate()
            return model_instance

        super(ModelType, self).__init__(validators=[validate_model] + validators, **kwargs)

    def __repr__(self):
        return object.__repr__(self)[:-1] + ' for %s>' % self.model_class

    def to_native(self, value, mapping=None, context=None):
        # We have already checked if the field is required. If it is None it
        # should continue being None
        if mapping is None:
            mapping = {}
        if value is None:
            return None
        if isinstance(value, self.model_class):
            return value

        if not isinstance(value, dict):
            raise ConversionError(
                u'Please use a mapping for this field or {0} instance instead of {1}.'.format(
                    self.model_class.__name__,
                    type(value).__name__))

        # partial submodels now available with import_data (ht ryanolson)
        model = self.model_class()
        return model.import_data(value, mapping=mapping, context=context,
                                 strict=self.strict)

    def export_loop(self, model_instance, field_converter,
                    role=None, print_none=False):
        """
        Calls the main `export_loop` implementation because they are both
        supposed to operate on models.
        """
        if isinstance(model_instance, self.model_class):
            model_class = model_instance.__class__
        else:
            model_class = self.model_class

        shaped = export_loop(model_class, model_instance,
                             field_converter,
                             role=role, print_none=print_none)

        if shaped and len(shaped) == 0 and self.allow_none():
            return shaped
        elif shaped:
            return shaped
        elif print_none:
            return shaped


class ListType(MultiType):

    def __init__(self, field, min_size=None, max_size=None, **kwargs):

        if not isinstance(field, BaseType):
            compound_field = kwargs.pop('compound_field', None)
            field = self.init_compound_field(field, compound_field, **kwargs)

        self.field = field
        self.min_size = min_size
        self.max_size = max_size

        validators = [self.check_length, self.validate_items] + kwargs.pop("validators", [])

        super(ListType, self).__init__(validators=validators, **kwargs)

    @property
    def model_class(self):
        return self.field.model_class

    def _force_list(self, value):
        if value is None or value == EMPTY_LIST:
            return []

        try:
            if isinstance(value, basestring):
                raise TypeError()

            if isinstance(value, dict):
                return [value[unicode(k)] for k in sorted(map(int, value.keys()))]

            return list(value)
        except TypeError:
            return [value]

    def to_native(self, value, context=None):
        items = self._force_list(value)

        return [self.field.to_native(item, context) for item in items]

    def check_length(self, value):
        list_length = len(value) if value else 0

        if self.min_size is not None and list_length < self.min_size:
            message = ({
                True: u'Please provide at least %d item.',
                False: u'Please provide at least %d items.',
            }[self.min_size == 1]) % self.min_size
            raise ValidationError(message)

        if self.max_size is not None and list_length > self.max_size:
            message = ({
                True: u'Please provide no more than %d item.',
                False: u'Please provide no more than %d items.',
            }[self.max_size == 1]) % self.max_size
            raise ValidationError(message)

    def validate_items(self, items):
        errors = []
        for item in items:
            try:
                self.field.validate(item)
            except ValidationError as exc:
                errors.append(exc.messages)
        if errors:
            raise ValidationError(errors)

    def export_loop(self, list_instance, field_converter,
                    role=None, print_none=False):
        """Loops over each item in the model and applies either the field
        transform or the multitype transform.  Essentially functions the same
        as `transforms.export_loop`.
        """
        data = []
        for value in list_instance:
            if hasattr(self.field, 'export_loop'):
                shaped = self.field.export_loop(value, field_converter,
                                                role=role)
                feels_empty = shaped and len(shaped) == 0
            else:
                shaped = field_converter(self.field, value)
                feels_empty = shaped is None

            # Print if we want empty or found a value
            if feels_empty and self.field.allow_none():
                data.append(shaped)
            elif shaped is not None:
                data.append(shaped)
            elif print_none:
                data.append(shaped)

        # Return data if the list contains anything
        if len(data) > 0:
            return data
        elif len(data) == 0 and self.allow_none():
            return data
        elif print_none:
            return data


class DictType(MultiType):

    def __init__(self, field, coerce_key=None, **kwargs):
        if not isinstance(field, BaseType):
            compound_field = kwargs.pop('compound_field', None)
            field = self.init_compound_field(field, compound_field, **kwargs)

        self.coerce_key = coerce_key or unicode
        self.field = field

        validators = [self.validate_items] + kwargs.pop("validators", [])

        super(DictType, self).__init__(validators=validators, **kwargs)

    @property
    def model_class(self):
        return self.field.model_class

    def to_native(self, value, safe=False, context=None):
        if value == EMPTY_DICT:
            value = {}

        value = value or {}

        if not isinstance(value, dict):
            raise ValidationError(u'Only dictionaries may be used in a DictType')

        return dict((self.coerce_key(k), self.field.to_native(v, context))
                    for k, v in iteritems(value))

    def validate_items(self, items):
        errors = {}
        for key, value in iteritems(items):
            try:
                self.field.validate(value)
            except ValidationError as exc:
                errors[key] = exc

        if errors:
            raise ValidationError(errors)

    def export_loop(self, dict_instance, field_converter,
                    role=None, print_none=False):
        """Loops over each item in the model and applies either the field
        transform or the multitype transform.  Essentially functions the same
        as `transforms.export_loop`.
        """
        data = {}

        for key, value in iteritems(dict_instance):
            if hasattr(self.field, 'export_loop'):
                shaped = self.field.export_loop(value, field_converter,
                                                role=role)
                feels_empty = shaped and len(shaped) == 0
            else:
                shaped = field_converter(self.field, value)
                feels_empty = shaped is None

            if feels_empty and self.field.allow_none():
                data[key] = shaped
            elif shaped is not None:
                data[key] = shaped
            elif print_none:
                data[key] = shaped

        if len(data) > 0:
            return data
        elif len(data) == 0 and self.allow_none():
            return data
        elif print_none:
            return data


class PolyModelType(MultiType):

    def __init__(self, model_classes, **kwargs):

        if isinstance(model_classes, type) and issubclass(model_classes, Model):
            self.model_classes = (model_classes,)
            allow_subclasses = True
        elif isinstance(model_classes, Iterable) \
          and not isinstance(model_classes, basestring):
            self.model_classes = tuple(model_classes)
            allow_subclasses = False
        else:
            raise Exception("The first argument to PolyModelType.__init__() "
                            "must be a model or an iterable.")

        validators = kwargs.pop("validators", [])
        self.strict = kwargs.pop("strict", True)
        self.claim_function = kwargs.pop("claim_function", None)
        self.allow_subclasses = kwargs.pop("allow_subclasses", allow_subclasses)

        def validate_model(model_instance):
            model_instance.validate()
            return model_instance

        MultiType.__init__(self, validators=[validate_model] + validators, **kwargs)

    def __repr__(self):
        return object.__repr__(self)[:-1] + ' for %s>' % str(self.model_classes)

    def is_allowed_model(self, model_instance):
        if self.allow_subclasses:
            if isinstance(model_instance, self.model_classes):
                return True
        else:
            if model_instance.__class__ in self.model_classes:
                return True
        return False

    def to_native(self, value, mapping=None, context=None):

        if mapping is None:
            mapping = {}
        if value is None:
            return None
        if self.is_allowed_model(value):
            return value
        if not isinstance(value, dict):
            if len(self.model_classes) > 1:
                instanceof_msg = 'one of: {}'.format(', '.join(
                    cls.__name__ for cls in self.model_classes))
            else:
                instanceof_msg = self.model_classes[0].__name__
            raise ConversionError(u'Please use a mapping for this field or '
                                    'an instance of {}'.format(instanceof_msg))

        model_class = self.find_model(value)
        model = model_class()
        return model.import_data(value, mapping=mapping, context=context,
                                 strict=self.strict)

    def find_model(self, data):
        """Finds the intended type by consulting potential classes or `claim_function`."""

        chosen_class = None
        if self.claim_function:
            chosen_class = self.claim_function(self, data)
        else:
            candidates = self.model_classes
            if self.allow_subclasses:
                candidates = itertools.chain.from_iterable(
                                 ([m] + m._subclasses for m in candidates))
            fallback = None
            matching_classes = []
            for cls in candidates:
                match = None
                if '_claim_polymorphic' in cls.__dict__:
                    match = cls._claim_polymorphic(data)
                elif not fallback: # The first model that doesn't define the hook
                    fallback = cls # can be used as a default if there's no match.
                if match:
                    matching_classes.append(cls)
            if not matching_classes and fallback:
                chosen_class = fallback
            elif len(matching_classes) == 1:
                chosen_class = matching_classes[0]
            else:
                raise Exception("Got ambiguous input for polymorphic field")
        if chosen_class:
            return chosen_class
        else:
            raise Exception("Input for polymorphic field did not match any model")

    def export_loop(self, model_instance, field_converter,
                    role=None, print_none=False):

        model_class = model_instance.__class__
        if not self.is_allowed_model(model_instance):
            raise Exception("Cannot export: {} is not an allowed type".format(model_class))

        shaped = export_loop(model_class, model_instance,
                             field_converter,
                             role=role, print_none=print_none)

        if shaped and len(shaped) == 0 and self.allow_none():
            return shaped
        elif shaped:
            return shaped
        elif print_none:
            return shaped


from ..models import Model
